from . import rbc_ukqcd as ru
from . import rbc_ukqcd_params as rup
from .jobs import *

# -----------------------------------------------------------------------------

@q.timer
def run_get_inverter(job_tag, traj, *, inv_type, get_gf, get_gt=None, get_eig=None):
    if None in [ get_gf, ]:
        return
    if get_gt is None:
        get_gt = lambda: None
    if get_eig is None:
        get_eig = lambda: None
    gf = get_gf()
    gt = get_gt()
    eig = get_eig()
    for inv_acc in [ 0, 1, 2, ]:
        ru.get_inv(gf, job_tag, inv_type, inv_acc, gt=gt, eig=eig)

# -----------------------------------------------------------------------------

@q.timer_verbose
def compute_prop_wsrc_1(
        job_tag, traj,
        *,
        gf, gt, eig,
        idx, tslice,
        inv_type, inv_acc,
        ):
    q.check_stop()
    q.check_time_limit()
    fname = q.get_fname()
    tag = f"tslice={tslice} ; type={inv_type} ; accuracy={inv_acc}"
    q.displayln_info(0, f"{fname}: {job_tag}/{traj} idx={idx} tag='{tag}'")
    inv = ru.get_inv(gf, job_tag, inv_type, inv_acc, gt=gt, eig=eig)
    total_site = q.Coordinate(get_param(job_tag, "total_site"))
    geo = q.Geometry(total_site)
    src = q.mk_wall_src(geo, tslice)
    sol = inv * src
    return sol

@q.timer
def compute_prop_wsrc_full(
        job_tag, traj,
        *,
        gf, gt, eig,
        idx, tslice,
        inv_type, inv_acc,
        sfw,
        ):
    fname = q.get_fname()
    tag = f"tslice={tslice} ; type={inv_type} ; accuracy={inv_acc}"
    if tag in sfw:
        return None
    prop = compute_prop_wsrc_1(
            job_tag, traj,
            gf=gf, gt=gt, eig=eig,
            idx=idx, tslice=tslice,
            inv_type=inv_type, inv_acc=inv_acc,
            )
    q.qnorm_field(prop).save_double(sfw, tag + " ; qnorm_field")
    prop.save_double(sfw, tag, skip_if_exist=True)
    sfw.flush()

@q.timer_verbose
def compute_prop_wsrc_full_all(job_tag, traj, *,
                               inv_type, gf, gt, wi, eig):
    inv_type_name_list = [ "light", "strange", ]
    inv_type_name = inv_type_name_list[inv_type]
    path_s = f"{job_tag}/prop-wsrc-full-{inv_type_name}/traj-{traj}"
    sfw = q.open_fields(get_save_path(path_s + ".acc"), "a", q.Coordinate([ 2, 2, 2, 4, ]))
    for inv_acc in [ 2, 1, ]:
        for p in wi:
            idx, tslice, inv_type_p, inv_acc_p=p
            if inv_type_p == inv_type and inv_acc_p == inv_acc:
                compute_prop_wsrc_full(
                        job_tag, traj,
                        gf=gf, gt=gt, eig=eig,
                        idx=idx, tslice=tslice, inv_type=inv_type, inv_acc=inv_acc,
                        sfw=sfw,
                        )
    sfw.close()
    q.qrename_info(get_save_path(path_s + ".acc"), get_save_path(path_s))

@q.timer(is_timer_fork=True)
def run_prop_wsrc_full(job_tag, traj, *, inv_type, get_gf, get_eig, get_gt, get_wi):
    fname = q.get_fname()
    if None in [ get_gf, get_gt, ]:
        return
    if get_eig is None:
        if inv_type == 0:
            return
        get_eig = lambda: None
    inv_type_name_list = [ "light", "strange", ]
    inv_type_name = inv_type_name_list[inv_type]
    path_f = f"{job_tag}/prop-wsrc-full-{inv_type_name}/traj-{traj}/geon-info.txt"
    path_s = f"{job_tag}/prop-wsrc-{inv_type_name}/traj-{traj}/geon-info.txt"
    if get_load_path(path_f) is not None:
        return
    if get_load_path(path_s) is not None:
        q.displayln_info(f"{fname}: {job_tag} {traj} {inv_type_name} already have sparse wsrc prop. Skip calculating the full prop.")
        return
    if q.obtain_lock(f"locks/{job_tag}-{traj}-{fname}-{inv_type_name}"):
        gf = get_gf()
        gt = get_gt()
        eig = get_eig()
        wi = get_wi()
        compute_prop_wsrc_full_all(job_tag, traj,
                                   inv_type=inv_type, gf=gf, gt=gt, wi=wi,
                                   eig=eig)
        q.release_lock()
        return [ f"{fname} {job_tag} {traj} {inv_type} done", ]

# -----------------------------------------------------------------------------

@q.timer
def avg_weight_from_prop_full(geo, prop_nf_dict):
    """
    prop_nf_dict[(inv_type, tslice,)] => prop_nf, prop_nf_glb_sum_tslice
    type(prop_nf) = q.FieldRealD
    type(prop_nf_glb_sum_tslice) = q.SelectedPointsRealD
    """
    fname = q.get_fname()
    inv_type_name_list = [ "light", "strange", ]
    total_site = geo.total_site
    n_samples = [ 0, 0, ]
    avg_nf_glb_sum_tslice = [ 0, 0, ]
    for k, v in prop_nf_dict.items():
        inv_type, tslice = k
        prop_nf, prop_nf_glb_sum_tslice = v
        gst = np.roll(prop_nf_glb_sum_tslice[:].ravel(), -tslice)
        n_samples[inv_type] += 1
        avg_nf_glb_sum_tslice[inv_type] += gst
    for inv_type in [ 0, 1, ]:
        inv_type_name = inv_type_name_list[inv_type]
        assert n_samples[inv_type] == total_site[3]
        avg_nf_glb_sum_tslice[inv_type] = avg_nf_glb_sum_tslice[inv_type] / (geo.total_volume / total_site[3] * n_samples[inv_type])
        q.displayln_info(-1, fname, "avg_nf_glb_sum_tslice", inv_type_name, avg_nf_glb_sum_tslice[inv_type])
    local_tsize = geo.local_site[3]
    local_t_start = geo.coor_node[3] * local_tsize
    local_t_end = local_t_start + local_tsize
    local_tslices = slice(local_t_start, local_t_end, 1)
    local_field_shape = tuple(reversed(geo.local_site.to_list()))
    f_weight_avg = []
    for inv_type in [ 0, 1, ]:
        f = q.FieldRealD(geo, 1)
        q.set_zero(f)
        f_weight_avg.append(f)
    for k, v in prop_nf_dict.items():
        inv_type, tslice = k
        prop_nf, prop_nf_glb_sum_tslice = v
        avg_gst = avg_nf_glb_sum_tslice[inv_type][:].ravel()
        avg_gst_local = np.roll(avg_gst, tslice)[local_tslices]
        weight = prop_nf[:].reshape(local_field_shape) / avg_gst_local[:, None, None, None]
        # q.displayln_info(q.avg_err(weight.ravel()))
        view = f_weight_avg[inv_type][:].reshape(local_field_shape)
        view += weight / n_samples[inv_type]
    f_weight_final = q.FieldRealD(geo, 1)
    q.set_zero(f_weight_final)
    f_weight_final[:] = 0.5
    for inv_type in [ 0, 1, ]:
        f_weight_final[:] += f_weight_avg[inv_type][:] / 4
    return f_weight_avg, f_weight_final

@q.timer
def make_fsel_from_weight(f_weight, f_rand_01, rate):
    """
    f_weight is expected to be averaged around 1.
    """
    fname = q.get_fname()
    geo = f_weight.geo
    fsel = q.FieldSelection(geo)
    sel = f_weight[:].ravel() * rate >= f_rand_01[:].ravel()
    val = np.rint(f_weight[:].ravel()[sel] * 10**8).astype(int)
    assert np.all(val >= 0)
    fsel[sel] = val
    fsel.update()
    q.displayln_info(-1, f"{fname}: rate = {rate} ; expect_num = {geo.total_volume * rate} ; actual_num = {q.glb_sum(fsel.n_elems)}")
    return fsel

@q.timer
def make_psel_from_weight(f_weight, f_rand_01, rate):
    """
    f_weight is expected to be averaged around 1.
    """
    fsel = make_fsel_from_weight(f_weight, f_rand_01, rate)
    psel = fsel.to_psel()
    return psel

@q.timer_verbose
def compute_f_weight_from_wsrc_prop_full(job_tag, traj, *,
        fn_f_weight,
        inv_type_list,
        inv_type_name_list,
        path_f_list,
        fn_f_weight_type_list,
        ):
    fname = q.get_fname()
    assert get_load_path(fn_f_weight) is None
    total_site = q.Coordinate(get_param(job_tag, "total_site"))
    geo = q.Geometry(total_site)
    prop_nf_dict = dict()
    idx = 0
    inv_acc = 1
    for inv_type in inv_type_list:
        path_f = path_f_list[inv_type]
        sfr = q.open_fields(get_load_path(path_f), "r")
        available_tags = sfr.list()
        q.displayln_info(0, f"available_tags={available_tags}")
        for tslice in range(total_site[3]):
            tag = f"tslice={tslice} ; type={inv_type} ; accuracy={inv_acc}"
            if tag not in sfr:
                continue
            q.displayln_info(0, f"{fname}: idx={idx} tag='{tag}'")
            idx += 1
            prop_nf = q.FieldRealD(geo, 1)
            q.set_zero(prop_nf)
            prop_nf.load_double(sfr, tag + " ; qnorm_field")
            prop_nf_dict[(inv_type, tslice,)] = prop_nf, prop_nf.glb_sum_tslice()
    assert idx >= 1
    f_weight_avg, f_weight_final = avg_weight_from_prop_full(geo, prop_nf_dict)
    for inv_type in [ 0, 1, ]:
        inv_type_name = inv_type_name_list[inv_type]
        f_weight_avg[inv_type].save_double(get_save_path(fn_f_weight_type_list[inv_type]))
        q.displayln_info(-1, fname, "field-selection-weight", inv_type_name, f_weight_avg[inv_type].glb_sum_tslice()[:].ravel())
    f_weight_final.save_double(get_save_path(fn_f_weight))
    q.displayln_info(-1, fname, "field-selection-weight final", f_weight_final.glb_sum_tslice()[:].ravel())

@q.timer(is_timer_fork=True)
def run_f_weight_from_wsrc_prop_full(job_tag, traj):
    """
    return get_f_weight
        f_weight = get_f_weight()
    Or if wsrc_prop_full is not available
    return None
    #
    `f_weight` is of type `q.FieldRealD(geo, 1)`.
    #
    get_f_weight = run_f_weight_from_wsrc_prop_full(job_tag, traj)
    """
    fname = q.get_fname()
    fn_f_weight = f"{job_tag}/field-selection-weight/traj-{traj}/weight.field"
    @q.lazy_call
    @q.timer_verbose
    def get_f_weight():
        f_weight = q.FieldRealD()
        total_bytes = f_weight.load_double(get_load_path(fn_f_weight))
        assert total_bytes > 0
        return f_weight
    ret = get_f_weight
    if get_load_path(fn_f_weight) is not None:
        return ret
    inv_type_list = [ 0, 1, ]
    inv_type_name_list = [ "light", "strange", ]
    path_f_list = [
            f"{job_tag}/prop-wsrc-full-{inv_type_name}/traj-{traj}/geon-info.txt"
            for inv_type_name in inv_type_name_list
            ]
    fn_f_weight_type_list = [
            f"{job_tag}/field-selection-weight/traj-{traj}/weight-{inv_type_name}.field"
            for inv_type_name in inv_type_name_list
            ]
    for inv_type in inv_type_list:
        if get_load_path(path_f_list[inv_type]) is None:
            inv_type_name = inv_type_name_list[inv_type]
            q.displayln_info(-1, f"WARNING: {fname}: {job_tag} {traj} {inv_type_name} full prop is not available yet.")
            return None
    if not q.obtain_lock(f"locks/{job_tag}-{traj}-{fname}"):
        return None
    compute_f_weight_from_wsrc_prop_full(
            job_tag, traj,
            fn_f_weight=fn_f_weight,
            inv_type_list=inv_type_list,
            inv_type_name_list=inv_type_name_list,
            path_f_list=path_f_list,
            fn_f_weight_type_list=fn_f_weight_type_list,
            )
    q.release_lock()
    return ret

@q.timer_verbose
def run_f_weight_uniform(job_tag, traj):
    """
    return get_f_weight
        f_weight = get_f_weight()
    #
    `f_weight` is of type `q.FieldRealD(geo, 1)`.
    #
    get_f_weight = run_f_weight_uniform(job_tag, traj)
    #
    Another option to obtain `get_f_weight` is using `run_f_weight_from_wsrc_prop_full(job_tag, traj)`.
    """
    fname = q.get_fname()
    fn_f_weight = f"{job_tag}/field-selection-weight/traj-{traj}/weight.field"
    @q.lazy_call
    @q.timer_verbose
    def get_f_weight():
        f_weight = q.FieldRealD()
        total_bytes = f_weight.load_double(get_load_path(fn_f_weight))
        assert total_bytes > 0
        return f_weight
    ret = get_f_weight
    if get_load_path(fn_f_weight) is not None:
        return ret
    if not q.obtain_lock(f"locks/{job_tag}-{traj}-{fname}"):
        return None
    total_site = q.Coordinate(get_param(job_tag, "total_site"))
    geo = q.Geometry(total_site)
    f_weight = q.FieldRealD(geo, 1)
    f_weight.set_unit()
    f_weight.save_double(get_save_path(fn_f_weight))
    q.release_lock()
    return ret

@q.timer_verbose
def run_f_weight_load(job_tag, traj):
    """
    return get_f_weight
        f_weight = get_f_weight()
    #
    `f_weight` is of type `q.FieldRealD(geo, 1)`.
    #
    get_f_weight = run_f_weight_load(job_tag, traj)
    #
    Just perform loading. Fail is data does not exist.
    """
    fname = q.get_fname()
    fn_f_weight = f"{job_tag}/field-selection-weight/traj-{traj}/weight.field"
    @q.lazy_call
    @q.timer_verbose
    def get_f_weight():
        f_weight = q.FieldRealD()
        total_bytes = f_weight.load_double(get_load_path(fn_f_weight))
        assert total_bytes > 0
        return f_weight
    ret = get_f_weight
    if get_load_path(fn_f_weight) is not None:
        return ret
    raise Exception(f"{fname}: 'fn_f_weight' does not exist.")

@q.timer_verbose
def run_f_rand_01(job_tag, traj):
    """
    return get_f_rand_01
        f_rand_01 = get_f_rand_01()
    #
    `f_rand_01` is of type `q.FieldRealD(geo, 1)`.
    #
    get_f_rand_01 = run_f_rand_01(job_tag, traj)
    """
    fname = q.get_fname()
    fn_f_rand_01 = f"{job_tag}/field-selection-weight/traj-{traj}/f-rand-01.field"
    @q.lazy_call
    @q.timer_verbose
    def get_f_rand_01():
        f_rand_01 = q.FieldRealD()
        total_bytes = f_rand_01.load_double(get_load_path(fn_f_rand_01))
        assert total_bytes > 0
        return f_rand_01
    ret = get_f_rand_01
    if get_load_path(fn_f_rand_01) is not None:
        return ret
    if not q.obtain_lock(f"locks/{job_tag}-{traj}-{fname}"):
        return None
    total_site = q.Coordinate(get_param(job_tag, "total_site"))
    geo = q.Geometry(total_site)
    seed = get_job_seed(job_tag)
    rs = q.RngState(f"{seed}-{traj}").split("run_sel_from_wsrc_prop_full").split("f_rand_01")
    f_rand_01 = q.FieldRealD(geo, 1)
    f_rand_01.set_rand(rs, 1.0, 0.0)
    f_rand_01.save_double(get_save_path(fn_f_rand_01))
    q.release_lock()
    return ret

@q.timer_verbose
def run_fsel_prob(job_tag, traj, *, get_f_rand_01, get_f_weight):
    """
    return get_fsel_prob
    #
        fsel_prob = get_fsel_prob()
        fsel = fsel_prob.fsel
    #
    One can set `get_f_rand_01=None` and `get_f_weight=None` if one intend to load existing data on `fsel_prob` instead of generating new `fsel_prob`.
    #
    `get_f_weight` is of type `lambda : q.FieldRealD(geo, 1)`.
    `get_f_rand_01` is of type `lambda : q.FieldRealD(geo, 1)`.
    #
    get_fsel_prob = run_fsel_prob(job_tag, traj, get_f_rand_01=get_f_rand_01, get_f_weight=get_f_weight)
    get_fsel = lambda : get_fsel_prob().fsel
    #
    For old data format, in which case the `fsel-prob.sfield` data is absent, we assume uniform probability AND will combine the `fsel` with `psel` obtained from `run_psel_prob` function.
    """
    fname = q.get_fname()
    fn_fsel = f"{job_tag}/field-selection/traj-{traj}.field"
    fn_fsel_prob = f"{job_tag}/field-selection-weight/traj-{traj}/fsel-prob.sfield"
    @q.lazy_call
    @q.timer_verbose
    def get_fsel_prob():
        fsel = q.FieldSelection()
        total_size = fsel.load(get_load_path(fn_fsel))
        assert total_size > 0
        fsel_prob = q.SelectedFieldRealD(fsel, 1)
        total_size = fsel_prob.load_double(get_load_path(fn_fsel_prob))
        assert total_size > 0
        return fsel_prob
    ret = get_fsel_prob
    if get_load_path(fn_fsel) is not None:
        if get_load_path(fn_fsel_prob) is not None:
            return ret
        else:
            q.displayln_info(f"{fname}: WARNING: field-selection exist but prob is not available. Assuming loading load old data format.")
            get_psel_prob = run_psel_prob(job_tag, traj, get_f_rand_01=get_f_rand_01, get_f_weight=get_f_weight)
            @q.lazy_call
            @q.timer_verbose
            def get_fsel_prob_old():
                fsel = q.FieldSelection()
                total_size = fsel.load(get_load_path(fn_fsel))
                assert total_size > 0
                psel = get_psel_prob().psel
                fsel.add_psel(psel)
                total_volume = fsel.total_site.volume()
                fsel_prob = q.SelectedFieldRealD(fsel, 1)
                fsel_prob[:] = q.glb_sum(len(fsel)) / total_volume
                return fsel_prob
            return get_fsel_prob_old
    if get_f_rand_01 is None:
        q.displayln_info(-1, f"{fname}: get_f_rand_01 is None")
        return None
    if get_f_weight is None:
        q.displayln_info(-1, f"{fname}: get_f_weight is None")
        return None
    if not q.obtain_lock(f"locks/{job_tag}-{traj}-{fname}"):
        return None
    fsel_rate = get_param(job_tag, "field_selection_fsel_rate")
    q.displayln_info(-1, fname, f"fsel_rate = {fsel_rate}")
    assert fsel_rate is not None
    assert get_load_path(fn_fsel) is None
    assert get_load_path(fn_fsel_prob) is None
    f_rand_01 = get_f_rand_01()
    f_weight = get_f_weight()
    fsel = make_fsel_from_weight(f_weight, f_rand_01, fsel_rate)
    fsel.save(get_save_path(fn_fsel))
    fsel_prob = q.SelectedFieldRealD(fsel, 1)
    fsel_prob @= f_weight
    fsel_prob[:] = np.minimum(1.0, fsel_prob[:] * fsel_rate)
    fsel_prob.save_double(get_save_path(fn_fsel_prob))
    q.release_lock()
    return ret

@q.timer_verbose
def run_psel_prob(job_tag, traj, *, get_f_rand_01, get_f_weight, tag=None):
    """
    return get_psel_prob
    #
        psel_prob = get_psel_prob()
        psel = psel_prob.psel
    #
    One can set `get_f_rand_01=None` and `get_f_weight=None` if one intend to load existing data on `psel_prob` instead of generating new `psel_prob`.
    #
    `get_f_weight` is of type `lambda : q.FieldRealD(geo, 1)`.
    `get_f_rand_01` is of type `lambda : q.FieldRealD(geo, 1)`.
    #
    tag can be "small", "median", "large", etc
    #
    get_psel_prob = run_psel_prob(job_tag, traj, get_f_rand_01=get_f_rand_01, get_f_weight=get_f_weight)
    get_psel = lambda : get_psel_prob().psel
    """
    fname = q.get_fname()
    if tag is None:
        fn_psel = f"{job_tag}/points-selection/traj-{traj}.lati"
        fn_psel_prob = f"{job_tag}/field-selection-weight/traj-{traj}/psel-prob.lat"
    else:
        fn_psel = f"{job_tag}/psel_{tag}/traj-{traj}/psel.lati"
        fn_psel_prob = f"{job_tag}/psel_{tag}/traj-{traj}/psel-prob.lat"
    total_site = q.Coordinate(get_param(job_tag, "total_site"))
    geo = q.Geometry(total_site)
    @q.lazy_call
    @q.timer_verbose
    def get_psel_prob():
        psel = q.PointsSelection()
        psel.load(get_load_path(fn_psel))
        psel_prob = q.SelectedPointsRealD(psel, 1)
        psel_prob.load(get_load_path(fn_psel_prob))
        return psel_prob
    ret = get_psel_prob
    if get_load_path(fn_psel) is not None:
        if get_load_path(fn_psel_prob) is not None:
            return ret
        else:
            q.displayln_info(f"{fname}: WARNING: points-selection exist but prob is not available. Assuming loading load old data format.")
            total_volume = total_site.volume()
            @q.lazy_call
            @q.timer_verbose
            def get_psel_prob_old():
                psel = q.PointsSelection()
                psel.load(get_load_path(fn_psel))
                psel_prob = q.SelectedPointsRealD(psel, 1)
                psel_prob[:] = len(psel) / total_volume
                return psel_prob
            return get_psel_prob_old
    if get_f_rand_01 is None:
        q.displayln_info(-1, f"{fname}: get_f_rand_01 is None")
        return None
    if get_f_weight is None:
        q.displayln_info(-1, f"{fname}: get_f_weight is None")
        return None
    if tag is None:
        if not q.obtain_lock(f"locks/{job_tag}-{traj}-{fname}"):
            return None
    else:
        if not q.obtain_lock(f"locks/{job_tag}-{traj}-{fname}-{tag}"):
            return None
    if tag is None:
        psel_rate = get_param(job_tag, "field_selection_psel_rate")
    else:
        psel_rate = get_param(job_tag, f"field_selection_psel_rate_{tag}")
    q.displayln_info(-1, fname, f"tag='{tag}' ; psel_rate={psel_rate} .")
    assert psel_rate is not None
    assert get_load_path(fn_psel) is None
    assert get_load_path(fn_psel_prob) is None
    f_rand_01 = get_f_rand_01()
    f_weight = get_f_weight()
    psel = make_psel_from_weight(f_weight, f_rand_01, psel_rate)
    psel.save(get_save_path(fn_psel))
    psel_prob = q.SelectedPointsRealD(psel, 1)
    psel_prob @= f_weight
    psel_prob[:] = np.minimum(1.0, psel_prob[:] * psel_rate)
    psel_prob.save(get_save_path(fn_psel_prob))
    q.release_lock()
    return ret

@q.timer
def run_fsel_from_fsel_prob(get_fsel_prob):
    if get_fsel_prob is None:
        return None
    return lambda : get_fsel_prob().fsel

@q.timer
def run_psel_from_psel_prob(get_psel_prob):
    if get_psel_prob is None:
        return None
    return lambda : get_psel_prob().psel

# -----------------------------------------------------------------------------

@q.timer_verbose
def run_fsel_prob_sub_sampling(
        job_tag, traj,
        *,
        sub_sampling_rate,
        get_fsel_prob,
        get_f_rand_01,
        get_f_weight,
        ):
    """
    `sub_sampling_rate == 1` implies complete sub-sampling.
    Approximately `sub_sampling_rate` portion of the original selection get selected.
    #
    If `get_f_weight is None` then use `fsel_prob * sub_sampling_rate` as prob to select.
    This is not exactly the same as use `f_weight`!
    """
    assert 1.0 >= sub_sampling_rate >= 0.0
    @q.lazy_call
    @q.timer_verbose
    def get_fsel_prob_sub():
        fname = q.get_fname()
        fsel_prob = get_fsel_prob()
        fsel = fsel_prob.fsel
        f_rand_01 = get_f_rand_01()
        geo = f_rand_01.geo
        sel = fsel[:] >= 0
        f_prob = q.FieldRealD(geo, 1)
        f_prob.set_zero()
        if get_f_weight is not None:
            fsel_rate = get_param(job_tag, "field_selection_fsel_rate")
            f_weight = get_f_weight()
            f_prob @= f_weight
            f_prob *= fsel_rate * sub_sampling_rate
        else:
            f_prob @= fsel_prob
            f_prob *= sub_sampling_rate
        sel_sub = f_prob[:].ravel() >= f_rand_01[:].ravel()
        assert np.all(sel_sub == sel_sub & (fsel[:] >= 0))
        fsel_sub = q.FieldSelection(geo)
        fsel_sub[sel_sub] = fsel[sel_sub]
        fsel_sub.update()
        original_num = q.glb_sum(fsel.n_elems)
        expect_num = original_num * sub_sampling_rate
        actual_num = q.glb_sum(fsel_sub.n_elems)
        q.displayln_info(-1, f"{fname}: sub_sampling_rate = {sub_sampling_rate} ; expect_num = {expect_num} ; actual_num = {actual_num}")
        fsel_prob_sub = q.SelectedFieldRealD(fsel_sub, 1)
        fsel_prob_sub @= f_prob
        return fsel_prob_sub
    return get_fsel_prob_sub

@q.timer_verbose
def run_psel_prob_sub_sampling(
        job_tag, traj,
        *,
        sub_sampling_rate,
        get_psel_prob,
        get_f_rand_01,
        get_f_weight,
        ):
    """
    `sub_sampling_rate == 1` implies complete sub-sampling.
    Approximately `sub_sampling_rate` portion of the original selection get selected.
    #
    If `get_f_weight is None` then use `psel_prob * sub_sampling_rate` as prob to select.
    This is not exactly the same as use `f_weight`!
    This is due to `psel_prob` is always less or equal to 1, `f_weight` does not have any upper limit.
    #
    if get_param(job_tag, "use_simple_psel_sub_sampling", default=False):
        Will simply use the first `sub_sampling_rate * original_num` of the original points.
    """
    assert 1.0 >= sub_sampling_rate >= 0.0
    @q.lazy_call
    @q.timer_verbose
    def get_psel_prob_sub():
        fname = q.get_fname()
        psel_prob = get_psel_prob()
        psel = psel_prob.psel
        original_num = psel.n_points
        total_site = psel.total_site
        use_simple_psel_sub_sampling = get_param(job_tag, "use_simple_psel_sub_sampling", default=False)
        if use_simple_psel_sub_sampling:
            prob_arr = psel_prob[:]
            avg = np.average(prob_arr)
            assert np.linalg.norm(prob_arr - avg) <= 1e-8
            expect_num = round(sub_sampling_rate * original_num)
            xg_arr_sub = psel.xg_arr[:actual_num]
            psel_sub = q.PointsSelection(total_site, xg_arr_sub)
            psel_prob_sub = q.SelectedFieldRealD(psel_sub, 1)
            psel_prob_sub[:] = avg * (expect_num / original_num)
            return psel_prob_sub
        f_rand_01 = get_f_rand_01()
        sp_rand_01 = q.SelectedPointsRealD(psel, 1)
        sp_rand_01 @= f_rand_01
        sp_prob = q.SelectedPointsRealD(psel, 1)
        sp_prob.set_zero()
        if get_f_weight is not None:
            psel_rate = get_param(job_tag, "field_selection_psel_rate")
            f_weight = get_f_weight()
            sp_prob @= f_weight
            sp_prob *= psel_rate * sub_sampling_rate
        else:
            sp_prob @= psel_prob
            sp_prob *= sub_sampling_rate
        sel_sub = sp_prob[:].ravel() >= sp_rand_01[:].ravel()
        xg_arr = psel.xg_arr
        xg_arr_sub = xg_arr[sel_sub]
        psel_sub = q.PointsSelection(total_site, xg_arr_sub)
        expect_num = original_num * sub_sampling_rate
        actual_num = psel_sub.n_points
        q.displayln_info(-1, f"{fname}: sub_sampling_rate = {sub_sampling_rate} ; expect_num = {expect_num} ; actual_num = {actual_num}")
        psel_prob_sub = q.SelectedPointsRealD(psel_sub, 1)
        psel_prob_sub @= sp_prob
        return psel_prob_sub
    return get_psel_prob_sub

# -----------------------------------------------------------------------------

@q.timer(is_timer_fork=True)
def run_psel_split(
        job_tag, traj,
        *,
        get_psel,
        num_piece,
        ):
    """
    Should set `num_piece` be in the form as 2^n.
    return psel_list
    #
    len(psel_list) == num_piece
    """
    assert num_piece >= 1
    fname = q.get_fname()
    path_psel_list = f"{job_tag}/points-selection-split/traj-{traj}/num-piece-{num_piece}"
    @q.lazy_call
    @q.timer(is_timer_fork=True)
    def get_psel_list():
        psel_list = [ q.PointsSelection() for idx in range(num_piece) ]
        load_path = get_load_path(path_psel_list + ".qar")
        assert load_path is not None
        for idx in range(num_piece):
            psel_list[idx].load(get_load_path(f"{path_psel_list}/idx-piece-{idx}.lati"))
        return psel_list
    ret = get_psel_list
    if get_load_path(path_psel_list + "/checkpoint.txt") is not None:
        return ret
    if get_psel is None:
        q.displayln_info(-1, f"{fname}: get_psel is None")
        return None
    if not q.obtain_lock(f"locks/{job_tag}-{traj}-{fname}"):
        return None
    seed = get_job_seed(job_tag)
    rs = q.RngState(f"{seed}-{traj}-{fname}")
    psel = get_psel()
    psel_list = q.psel_split_n_that_increase_separation(psel, num_piece, rs=rs)
    assert len(psel_list) == num_piece
    if 0 == q.get_id_node():
        qar = q.open_qar(get_save_path(path_psel_list + ".qar"), "w")
        for idx in range(num_piece):
            fn = f"idx-piece-{idx}.lati"
            qar.write(fn, "", psel_list[idx].save_str(), skip_if_exist=True)
        qar.write("checkpoint.txt", "", "", skip_if_exist=True)
    q.sync_node()
    q.release_lock()
    return ret

@q.timer(is_timer_fork=True)
def run_fsel_split(
        job_tag, traj,
        *,
        get_fsel,
        num_piece,
        ):
    """
    Should set `num_piece` be in the form as 2^n.
    return psel_list
    #
    len(psel_list) == num_piece
    """
    assert num_piece >= 1
    fname = q.get_fname()
    path_psel_list = f"{job_tag}/field-selection-split/traj-{traj}/num-piece-{num_piece}"
    @q.lazy_call
    @q.timer(is_timer_fork=True)
    def get_psel_list():
        psel_list = [ q.PointsSelection() for idx in range(num_piece) ]
        load_path = get_load_path(path_psel_list + ".qar")
        assert load_path is not None
        for idx in range(num_piece):
            psel_list[idx].load(get_load_path(f"{path_psel_list}/idx-piece-{idx}.lati"))
        return psel_list
    ret = get_psel_list
    if get_load_path(path_psel_list + "/checkpoint.txt") is not None:
        return ret
    if get_fsel is None:
        q.displayln_info(-1, f"{fname}: get_fsel is None")
        return None
    if not q.obtain_lock(f"locks/{job_tag}-{traj}-{fname}"):
        return None
    seed = get_job_seed(job_tag)
    rs = q.RngState(f"{seed}-{traj}-{fname}")
    psel = get_fsel().to_psel()
    psel_list = q.psel_split_n_that_increase_separation(psel, num_piece, rs=rs)
    assert len(psel_list) == num_piece
    if 0 == q.get_id_node():
        qar = q.open_qar(get_save_path(path_psel_list + ".qar"), "w")
        for idx in range(num_piece):
            fn = f"idx-piece-{idx}.lati"
            qar.write(fn, "", psel_list[idx].save_str(), skip_if_exist=True)
        qar.write("checkpoint.txt", "", "", skip_if_exist=True)
    q.sync_node()
    q.release_lock()
    return ret

# -----------------------------------------------------------------------------

@q.timer_verbose
def save_prop_wsrc_sparse(job_tag, traj, *, load_prop, tslice, inv_type, inv_acc, sfw, qar_sp, psel, fsel):
    """
    `load_prop()` is the wall source propagator, as if generated in a Coulomb gauge fixed gauge configuration.
    `tslice` is the integer indicate the source time slice.
    `inv_type` is `0` (light quark) or `1` (strange quark) or `2` (charm quark)
    `inv_acc` for wall source typically be integer `1` (sloppy accuracy) or `2` (exact accuracy).
    """
    fname = q.get_fname()
    tag = f"tslice={tslice} ; type={inv_type} ; accuracy={inv_acc}"
    if tag in sfw:
        if 0 == q.get_id_node():
            assert f"{tag}.lat" in qar_sp
            assert f"{tag} ; wsnk.lat" in qar_sp
        q.displayln_info(0, f"{fname}: tag='{tag}' of '{job_tag}/{traj}' already saved. Skipping.")
        return
    prop = load_prop()
    if prop is None:
        return
    q.displayln_info(-1, f"{fname}: write wsrc prop tag='{tag}' of '{job_tag}/{traj}'")
    s_prop = q.SelProp(fsel)
    ps_prop = q.PselProp(psel)
    q.set_zero(s_prop)
    q.set_zero(ps_prop)
    s_prop @= prop
    ps_prop @= prop
    ps_prop_ws = prop.glb_sum_tslice()
    qar_sp.write(f"{tag}.lat", "", ps_prop.save_str(), skip_if_exist=True)
    qar_sp.write(f"{tag} ; wsnk.lat", "", ps_prop_ws.save_str(), skip_if_exist=True)
    s_prop.save_float_from_double(sfw, tag, skip_if_exist=True)
    qar_sp.flush()
    sfw.flush()

@q.timer(is_timer_fork=True)
def run_prop_wsrc_sparse(
        job_tag, traj,
        *,
        inv_type, get_gf, get_gt, get_eig, get_psel, get_fsel, get_wi,
        ):
    fname = q.get_fname()
    if None in [ get_gt, get_psel, get_fsel, get_wi, ]:
        return
    inv_type_name_list = [ "light", "strange", ]
    inv_type_name = inv_type_name_list[inv_type]
    path_f = f"{job_tag}/prop-wsrc-full-{inv_type_name}/traj-{traj}/geon-info.txt"
    path_s = f"{job_tag}/prop-wsrc-{inv_type_name}/traj-{traj}"
    path_sp = f"{job_tag}/psel-prop-wsrc-{inv_type_name}/traj-{traj}"
    is_performing_inversion = get_param(job_tag, "is_performing_inversion_if_no_full_prop_available", default=False)
    if get_load_path(path_f) is None:
        if not is_performing_inversion:
            q.displayln_info(f"WARNING: {fname}: {job_tag} {traj} {inv_type_name} full prop is not available yet.")
            return
    if get_load_path(path_s + "/geon-info.txt") is not None:
        assert get_load_path(path_sp + "/checkpoint.txt") is not None
        return
    if not q.obtain_lock(f"locks/{job_tag}-{traj}-{fname}-{inv_type_name}"):
        return
    total_site = q.Coordinate(get_param(job_tag, "total_site"))
    geo = q.Geometry(total_site)
    gt = get_gt()
    fsel = get_fsel()
    psel = get_psel()
    wi = get_wi()
    if get_load_path(path_f) is not None:
        sfr = q.open_fields(get_load_path(path_f), "r")
        available_tags = sfr.list()
        q.displayln_info(0, f"available_tags={available_tags}")
    else:
        assert is_performing_inversion
        sfr = None
        if None in [ get_gf, get_eig, ]:
            assert False
        gf = get_gf()
        eig = get_eig()
    sfw = q.open_fields(get_save_path(path_s + ".acc"), "a", q.Coordinate([ 2, 2, 2, 4, ]))
    qar_sp = q.open_qar_info(get_save_path(path_sp + ".qar"), "a")
    for idx, tslice, inv_type_wi, inv_acc in wi:
        if inv_type_wi != inv_type:
            continue
        def load_prop():
            tag = f"tslice={tslice} ; type={inv_type} ; accuracy={inv_acc}"
            prop = q.Prop(geo)
            if sfr is not None:
                if tag not in sfr:
                    raise Exception(f"{tag} not in {sfr.list()}")
                prop.load_double(sfr, tag)
            else:
                prop = compute_prop_wsrc_1(
                        job_tag, traj,
                        gf=gf, gt=gt, eig=eig,
                        idx=idx, tslice=tslice,
                        inv_type=inv_type, inv_acc=inv_acc,
                        )
            return prop
        save_prop_wsrc_sparse(
                job_tag, traj,
                load_prop=load_prop,
                tslice=tslice, inv_type=inv_type, inv_acc=inv_acc,
                sfw=sfw, qar_sp=qar_sp,
                psel=psel, fsel=fsel,
                )
    sfw.close()
    qar_sp.write("checkpoint.txt", "", "", skip_if_exist=True)
    qar_sp.flush()
    qar_sp.close()
    q.qrename_info(get_save_path(path_s + ".acc"), get_save_path(path_s))
    q.release_lock()
    return [ f"{fname} {job_tag} {traj} {inv_type} done", ]

# -----------------------------------------------------------------------------

@q.timer
def calc_hvp_sum_tslice(chvp_16):
    """
    return ld_hvp_ts
    #
    ld_hvp_ts[t_dir, t, mu, nu]
    #
    t_dir in [ "x", "y", "z", "t", ]
    t in range(t_size)
    mu in range(4)
    nu in range(4)
    #
    `mu` for sink and `nu` for source
    t_size = max(total_site)
    #
    arr_for_t_dir.shape == (t_size, 4, 4,)
    #
    chvp_16 is a hvp field (from q.contract_chvp_16)
    """
    total_site = chvp_16.total_site
    t_size = max(total_site)
    ld_hvp_ts = q.mk_lat_data([
        [ "t_dir", 4, [ "x", "y", "z", "t", ], ],
        [ "t", t_size, ],
        [ "mu", 4, ],
        [ "nu", 4, ],
        ])
    ld_hvp_ts.set_zero()
    ld_arr = ld_hvp_ts[:]
    assert ld_arr.shape == (4, t_size, 4, 4,)
    assert ld_arr.dtype == np.complex128
    for t_dir in range(4):
        chvp_16_ts = chvp_16.glb_sum_tslice(t_dir=t_dir)
        arr = chvp_16_ts.to_numpy()
        t_size = arr.shape[0]
        ld_arr[t_dir, :t_size] = arr.reshape(t_size, 4, 4)
    return ld_hvp_ts

@q.timer
def compute_prop_psrc_hvp_contract(
        job_tag, traj, xg_src, inv_type, inv_acc,
        *,
        prop, tag, sfw_hvp, qar_hvp_ts):
    """
    # chvp_16.get_elem(x, mu * 4 + nu) is complex
    # (1) mu is the sink polarization and nu is the src polarization
    # (2) hvp field is simply the trace of the products of gamma matrix and propagators.
    #     It does not include the any minus sign (e.g. The minus sign due to the loop).
    """
    if (sfw_hvp is None) and (qar_hvp_ts is None):
        return
    assert isinstance(xg_src, q.Coordinate)
    chvp_16 = q.contract_chvp_16(prop, prop)
    ld_hvp_ts = calc_hvp_sum_tslice(chvp_16)
    if qar_hvp_ts is not None:
        qar_hvp_ts.write(f"{tag}.lat", "", ld_hvp_ts.save_str(), skip_if_exist=True)
        qar_hvp_ts.flush()
    if sfw_hvp is not None:
        chvp_16.save_float_from_double(sfw_hvp, tag, skip_if_exist=True)
        sfw_hvp.flush()

# -----------------------------------------------------------------------------

@q.timer_verbose
def compute_prop_2(inv, src, *, tag, sfw, qar_sp, psel, fsel,
                   f_rand_01, fsel_psrc_prop_norm_threshold, gt):
    sol = inv * src
    sp_sol = q.PselProp(psel)
    sp_sol @= sol
    qar_sp.write(f"{tag}.lat", "", sp_sol.save_str(), skip_if_exist=True)
    sol_gt = gt * sol
    sol_ws = sol_gt.glb_sum_tslice()
    qar_sp.write(f"{tag} ; wsnk.lat", "", sol_ws.save_str(), skip_if_exist=True)
    sol_ps_sel_prob = q.qnorm_field(sol)
    sol_ps_sel_prob *= 1.0 / fsel_psrc_prop_norm_threshold
    sol_ps_sel_prob[:] = np.minimum(1.0, sol_ps_sel_prob[:])
    ps_sel = f_rand_01[:, 0] <= sol_ps_sel_prob[:, 0]
    fsel_ps = q.FieldSelection(fsel.geo)
    fsel_ps[ps_sel] = 0
    fsel_ps.update()
    fsel_combine = fsel_ps.copy()
    fsel_combine.add_fsel(fsel)
    num_fsel = q.glb_sum(fsel.n_elems)
    num_fsel_ps = q.glb_sum(fsel_ps.n_elems)
    num_fsel_combine = q.glb_sum(fsel_combine.n_elems)
    q.displayln_info(0, f"compute_prop_psrc: tag='{tag}' ; num_fsel={num_fsel} ; num_fsel_ps={num_fsel_ps} ; num_fsel_combine={num_fsel_combine}")
    s_sol_ps_sel_prob = q.SelectedFieldRealD(fsel_ps)
    s_sol_ps_sel_prob @= sol_ps_sel_prob
    s_sol_ps_sel_prob.save_double(sfw, f"{tag} ; fsel-prob-psrc-prop", skip_if_exist=True)
    s_sol = q.SelProp(fsel_combine)
    s_sol @= sol
    s_sol.save_float_from_double(sfw, tag, skip_if_exist=True)
    qar_sp.flush()
    sfw.flush()
    return sol

def mk_psrc_tag(xg, inv_type, inv_acc):
    xg_str = f"({xg[0]},{xg[1]},{xg[2]},{xg[3]})"
    tag = f"xg={xg_str} ; type={inv_type} ; accuracy={inv_acc}"
    return tag

@q.timer
def compute_prop_psrc(job_tag, traj, xg_src, inv_type, inv_acc, *,
        idx, gf, gt, sfw, qar_sp, psel, fsel, f_rand_01, sfw_hvp, qar_hvp_ts,
        eig):
    assert isinstance(xg_src, q.Coordinate)
    tag = mk_psrc_tag(xg_src, inv_type, inv_acc)
    if (tag in sfw) and (sfw_hvp is None or (tag in sfw_hvp)):
        assert f"{tag} ; fsel-prob-psrc-prop" in sfw
        if 0 == q.get_id_node():
            assert f"{tag}.lat" in qar_sp
            assert f"{tag} ; wsnk.lat" in qar_sp
        if qar_hvp_ts is not None:
            assert qar_hvp_ts.has_regular_file(f"{tag}.lat")
        return None
    q.check_stop()
    q.check_time_limit()
    q.displayln_info(f"compute_prop_psrc: {job_tag} idx={idx} tag='{tag}'")
    inv = ru.get_inv(gf, job_tag, inv_type, inv_acc, eig=eig)
    total_site = q.Coordinate(get_param(job_tag, "total_site"))
    fsel_psrc_prop_norm_threshold = get_param(job_tag, "field_selection_fsel_psrc_prop_norm_threshold")
    geo = q.Geometry(total_site)
    src = q.mk_point_src(geo, xg_src)
    prop = compute_prop_2(
            inv, src, tag=tag, sfw=sfw, qar_sp=qar_sp,
            psel=psel, fsel=fsel,
            f_rand_01=f_rand_01,
            fsel_psrc_prop_norm_threshold=fsel_psrc_prop_norm_threshold,
            gt=gt)
    compute_prop_psrc_hvp_contract(
            job_tag, traj, xg_src, inv_type, inv_acc,
            prop=prop, tag=tag, sfw_hvp=sfw_hvp, qar_hvp_ts=qar_hvp_ts)

@q.timer_verbose
def compute_prop_psrc_all(job_tag, traj, *,
                          inv_type, gf, gt, psel, fsel, f_rand_01, eig):
    inv_type_name_list = [ "light", "strange", ]
    inv_type_name = inv_type_name_list[inv_type]
    path_s = f"{job_tag}/prop-psrc-{inv_type_name}/traj-{traj}"
    path_s_hvp = f"{job_tag}/hvp-psrc-{inv_type_name}/traj-{traj}"
    path_hvp_ts = f"{job_tag}/hvp-sum-tslice-psrc-{inv_type_name}/traj-{traj}"
    path_sp = f"{job_tag}/psel-prop-psrc-{inv_type_name}/traj-{traj}"
    sfw = q.open_fields(get_save_path(path_s + ".acc"), "a", q.Coordinate([ 2, 2, 2, 4, ]))
    is_saving_hvp = get_param(job_tag, "run_prop_psrc", "is_saving_hvp", default=True)
    is_saving_hvp_ts = get_param(job_tag, "run_prop_psrc", "is_saving_hvp_ts", default=True)
    if is_saving_hvp:
        sfw_hvp = q.open_fields(get_save_path(path_s_hvp + ".acc"), "a", q.Coordinate([ 2, 2, 2, 4, ]))
    else:
        sfw_hvp = None
    qar_sp = q.open_qar_info(get_save_path(path_sp + ".qar"), "a")
    if is_saving_hvp_ts:
        qar_hvp_ts = q.open_qar_info(get_save_path(path_hvp_ts + ".qar"), "a")
    else:
        qar_hvp_ts = None
    def comp(idx, xg_src, inv_acc):
        compute_prop_psrc(job_tag, traj, xg_src, inv_type, inv_acc,
                idx=idx, gf=gf, gt=gt, sfw=sfw, qar_sp=qar_sp,
                psel=psel, fsel=fsel, f_rand_01=f_rand_01,
                sfw_hvp=sfw_hvp, qar_hvp_ts=qar_hvp_ts,
                eig=eig)
    prob1 = get_param(job_tag, "prob_acc_1_psrc")
    prob2 = get_param(job_tag, "prob_acc_2_psrc")
    seed = get_job_seed(job_tag)
    rs = q.RngState(f"seed {seed} {traj}").split(f"compute_prop_psrc_all(ama)")
    for idx, xg_src in enumerate(psel):
        r = rs.split(f"{xg_src.to_tuple()}").u_rand_gen()
        assert 0 <= r and r <= 1
        comp(idx, xg_src, inv_acc=0)
        if r <= prob1:
            comp(idx, xg_src, inv_acc=1)
        if r <= prob2:
            comp(idx, xg_src, inv_acc=2)
    sfw.close()
    if sfw_hvp is not None:
        sfw_hvp.close()
    qar_sp.write("checkpoint.txt", "", "", skip_if_exist=True)
    qar_sp.flush()
    qar_sp.close()
    if qar_hvp_ts is not None:
        qar_hvp_ts.write("checkpoint.txt", "", "", skip_if_exist=True)
        qar_hvp_ts.flush()
        qar_hvp_ts.close()
    q.qrename_info(get_save_path(path_s + ".acc"), get_save_path(path_s))
    if sfw_hvp is not None:
        q.qrename_info(get_save_path(path_s_hvp + ".acc"), get_save_path(path_s_hvp))

@q.timer(is_timer_fork=True)
def run_prop_psrc(job_tag, traj, *, inv_type, get_gf, get_eig, get_gt, get_psel, get_fsel, get_f_rand_01):
    fname = q.get_fname()
    if None in [ get_gf, get_gt, get_psel, get_fsel, get_f_rand_01, ]:
        return
    if get_eig is None:
        if inv_type == 0:
            return
        get_eig = lambda: None
    inv_type_name_list = [ "light", "strange", ]
    inv_type_name = inv_type_name_list[inv_type]
    if get_load_path(f"{job_tag}/prop-psrc-{inv_type_name}/traj-{traj}/geon-info.txt") is not None:
        return
    if q.obtain_lock(f"locks/{job_tag}-{traj}-{fname}-{inv_type_name}"):
        gf = get_gf()
        gt = get_gt()
        eig = get_eig()
        psel = get_psel()
        fsel = get_fsel()
        f_rand_01 = get_f_rand_01()
        assert fsel.is_containing(psel)
        compute_prop_psrc_all(job_tag, traj,
                              inv_type=inv_type, gf=gf, gt=gt,
                              psel=psel, fsel=fsel,
                              f_rand_01=f_rand_01,
                              eig=eig)
        q.release_lock()
        return [ f"{fname} {job_tag} {traj} {inv_type} done", ]

# -----------------------------------------------------------------------------

@q.timer_verbose
def compute_hvp_average(job_tag, traj, *, inv_type, psel_prob, data_path, geo):
    fname = q.get_fname()
    psel = psel_prob.psel
    psel_prob_arr = psel_prob[:].ravel()
    hvp_average = q.FieldComplexD(geo, 16)
    hvp_average.set_zero()
    rel_acc_list = [ 0, 1, 2, ]
    prob_list = [ get_param(job_tag, f"prob_acc_{inv_acc}_psrc") for inv_acc in rel_acc_list ]
    sfr = q.open_fields(data_path, "r")
    tags = sfr.list()
    for p_idx in range(len(psel)):
        xg_src = q.Coordinate(psel[p_idx])
        prob_src = psel_prob_arr[p_idx]
        val_list = []
        for inv_acc in rel_acc_list:
            tag = mk_psrc_tag(xg_src, inv_type, inv_acc)
            if tag not in tags:
                val_list.append(None)
            else:
                chvp_16 = q.FieldComplexD(geo, 16)
                chvp_16.load_double_from_float(sfr, tag)
                val_list.append(chvp_16)
        assert val_list[0] is not None
        ama_val = q.mk_ama_val(val_list[0], xg_src.to_tuple(), val_list, rel_acc_list, prob_list)
        hvp = q.ama_extract(ama_val).shift(-xg_src)
        hvp *= 1 / prob_src
        hvp_average += hvp
    sfr.close()
    hvp_average *= 1 / geo.total_volume
    return hvp_average

@q.timer(is_timer_fork=True)
def run_hvp_average(job_tag, traj, *, inv_type, get_psel_prob):
    """
    return get_hvp_average()
    save hvp_average.field in single precision.
    hvp_average.get_elem(x, mu * 4 + nu) is complex
    #
    (1) mu is the sink polarization and nu is the src polarization
    (2) hvp field is simply the trace of the products of gamma matrix and propagators.
        It does not include the any minus sign (e.g. The minus sign due to the loop).
    """
    fname = q.get_fname()
    inv_type_name_list = [ "light", "strange", ]
    inv_type_name = inv_type_name_list[inv_type]
    fn = f"{job_tag}/hvp-average/traj-{traj}/hvp_average_{inv_type_name}.field"
    total_site = q.Coordinate(get_param(job_tag, "total_site"))
    geo = q.Geometry(total_site)
    @q.lazy_call
    def load():
        hvp_average = q.FieldComplexD(geo, 16)
        hvp_average.load_double_from_float(get_load_path(fn))
        return hvp_average
    if get_load_path(fn) is not None:
        return load
    data_fn = f"{job_tag}/hvp-psrc-{inv_type_name}/traj-{traj}/geon-info.txt"
    data_path = get_load_path(data_fn)
    if data_path is None:
        q.displayln_info(f"{fname}: '{data_fn}' not available.")
        return
    q.check_stop()
    q.check_time_limit()
    if not q.obtain_lock(f"locks/{job_tag}-{traj}-{fname}-{inv_type_name}"):
        return
    psel_prob = get_psel_prob()
    hvp_average = compute_hvp_average(job_tag, traj, inv_type=inv_type, psel_prob=psel_prob, geo=geo, data_path=data_path)
    hvp_average.save_float_from_double(get_save_path(fn))
    q.displayln_info(f"{fname}: {job_tag} {traj} {inv_type} num_fields={len(psel_prob.psel)}")
    q.release_lock()
    return load

# -----------------------------------------------------------------------------

@q.timer(is_timer_fork=True)
def run_field_rand_u1_dict(
        job_tag, traj,
        ):
    """
    return get_field_rand_u1_dict
    #
    get_field_rand_u1_dict()["fsel-src"] => q.FieldComplexD
    get_field_rand_u1_dict()["fsel-src-dag"] => q.FieldComplexD
    get_field_rand_u1_dict()["psel-src"] => q.FieldComplexD
    get_field_rand_u1_dict()["psel-src-dag"] => q.FieldComplexD
    """
    fname = q.get_fname()
    path = f"{job_tag}/field-rand-u1/traj-{traj}"
    @q.lazy_call
    @q.timer_verbose
    def get_field_rand_u1_dict():
        d = dict()
        for psel_list_type in [ "fsel", "psel" ]:
            fn = get_load_path(f"{path}/{psel_list_type}-src.field")
            fn_dag = get_load_path(f"{path}/{psel_list_type}-src-dag.field")
            assert fn is not None
            assert fn_dag is not None
            f = q.FieldComplexD()
            f_dag = q.FieldComplexD()
            f.load_double(fn)
            f_dag.load_double(fn_dag)
            d[f"{psel_list_type}-src"] = f
            d[f"{psel_list_type}-src-dag"] = f_dag
        return d
    ret = get_field_rand_u1_dict
    if get_load_path(f"{path}/checkpoint.txt"):
        return ret
    total_site = q.Coordinate(get_param(job_tag, "total_site"))
    geo = q.Geometry(total_site)
    seed = get_job_seed(job_tag)
    rs_rand_u1 = q.RngState(f"seed {seed} {traj}").split(f"compute_prop_rand_sparse_u1_src(rand_u1)")
    for psel_list_type in [ "fsel", "psel" ]:
        fu1 = q.mk_rand_vol_u1(geo, 1, rs_rand_u1.split(f"{psel_list_type}"))
        fu1_dag = q.FieldComplexD(geo, 1)
        fu1_dag[:] = fu1[:].conj()
        fu1.save_double(get_save_path(f"{path}/{psel_list_type}-src.field"))
        fu1_dag.save_double(get_save_path(f"{path}/{psel_list_type}-src-dag.field"))
        if is_test():
            q.json_results_append(f"{fname}: {psel_list_type} fu1", q.get_data_sig_arr(fu1, q.RngState(), 4), 1e-15)
            q.json_results_append(f"{fname}: {psel_list_type} fu1_dag", q.get_data_sig_arr(fu1_dag, q.RngState(), 4), 1e-15)
    q.qtouch_info(get_save_path(f"{path}/checkpoint.txt"), "")
    return ret

@q.timer(is_timer_fork=True)
def run_prop_sparse_rand_u1_src(
        job_tag, traj,
        *,
        inv_type,
        get_gf,
        get_psel,
        get_fsel,
        get_field_rand_u1_dict,
        get_psel_list=None,
        get_fsel_psel_list=None,
        get_eig=None,
        ):
    """
    fsel should contain psel
    Should set one (and only one) of the `get_psel_list` and `get_fsel_psel_list`.
    """
    fname = q.get_fname()
    if None in [ get_gf, get_psel, get_fsel, get_field_rand_u1_dict, ]:
        return
    if (get_psel_list is None) and (get_fsel_psel_list is None):
        return None
    if get_eig is None:
        if inv_type == 0:
            return None
        else:
            get_eig = lambda: None
    if get_psel_list is not None:
        assert get_fsel_psel_list is None
        psel_list_type = "psel"
    elif get_fsel_psel_list is not None:
        psel_list_type = "fsel"
    else:
        assert False
    quark_flavor_list = get_param(job_tag, "quark_flavor_list")
    quark_flavor = quark_flavor_list[inv_type]
    path_s = f"{job_tag}/prop-rand-u1-{psel_list_type}-sparse-{quark_flavor}/traj-{traj}"
    path_sp = f"{job_tag}/psel-prop-rand-u1-{psel_list_type}-sparse-{quark_flavor}/traj-{traj}"
    if get_load_path(path_s + "/geon-info.txt") is not None:
        assert get_load_path(path_sp + "/checkpoint.txt") is not None
        return
    if not q.obtain_lock(f"locks/{job_tag}-{traj}-{fname}-{psel_list_type}-{quark_flavor}"):
        return
    gf = get_gf()
    geo = gf.geo
    fsel = get_fsel()
    psel = get_psel()
    field_rand_u1_dict = get_field_rand_u1_dict()
    eig = get_eig()
    if psel_list_type == "psel":
        assert get_psel_list is not None
        assert get_fsel_psel_list is None
        psel_list = get_psel_list()
        fu1 = field_rand_u1_dict["psel-src"]
        fu1_dag = field_rand_u1_dict["psel-src-dag"]
    elif psel_list_type == "fsel":
        assert get_psel_list is None
        assert get_fsel_psel_list is not None
        psel_list = get_fsel_psel_list()
        fu1 = field_rand_u1_dict["fsel-src"]
        fu1_dag = field_rand_u1_dict["fsel-src-dag"]
    else:
        assert False
    total_site = geo.total_site
    gt = None
    prob_acc_1_rand_sparse_u1 = get_param(job_tag, "prob_acc_1_rand_u1_sparse")
    prob_acc_2_rand_sparse_u1 = get_param(job_tag, "prob_acc_2_rand_u1_sparse")
    sfw = q.open_fields(get_save_path(path_s + ".acc"), "a", q.Coordinate([ 2, 2, 2, 4, ]))
    qar_sp = q.open_qar_info(get_save_path(path_sp + ".qar"), "a")
    fu1_from_sp = q.FieldComplexD(geo, 1)
    seed = get_job_seed(job_tag)
    rs_ama = q.RngState(f"seed {seed} {traj}").split(f"compute_prop_rand_u1(ama)")
    @q.timer
    def compute_and_save(idx_psel, is_dagger, inv_acc):
        nonlocal fu1_from_sp
        assert 0 <= idx_psel < len(psel_list)
        assert is_dagger in [ 0, 1, ]
        tag = f"idx_psel={idx_psel} ; is_dagger={is_dagger} ; type={inv_type} ; accuracy={inv_acc}"
        if tag in sfw:
            if q.get_id_node() == 0:
                assert f"{tag} ; psel_rand_src.lati" in qar_sp
                assert f"{tag} ; psel_rand_src ; fu1.lat" in qar_sp
                assert f"{tag} ; psel_rand_src ; prop.lat" in qar_sp
                assert f"{tag}.lat" in qar_sp
            return
        if q.get_id_node() == 0:
            if f"{tag}.lat" in qar_sp:
                q.displayln_info(-1, f"WARNING: {fname}: '{tag}.lat' already exist in {qar_sp.path()}")
        if f"{tag} ; fu1" in sfw:
            q.displayln_info(-1, f"WARNING: {fname}: '{tag} ; fu1' already exist in {sfw.path()}")
        q.check_stop()
        q.check_time_limit()
        psel_rand_src = psel_list[idx_psel]
        sp_fu1 = q.SelectedPointsComplexD(psel_rand_src, 1)
        sp_fu1.set_zero()
        if is_dagger == 0:
            sp_fu1 @= fu1
        elif is_dagger == 1:
            sp_fu1 @= fu1_dag
        else:
            assert False
        fu1_from_sp.set_zero()
        fu1_from_sp @= sp_fu1
        prop_src = q.mk_rand_vol_u1_src(fu1_from_sp)
        inv = ru.get_inv(gf, job_tag, inv_type, inv_acc, gt=gt, eig=eig)
        prop_sol = inv * prop_src
        s_prop = q.SelProp(fsel)
        ps_prop = q.PselProp(psel)
        ps_prop_prs = q.PselProp(psel_rand_src)
        s_prop @= prop_sol
        ps_prop @= prop_sol
        ps_prop_prs @= prop_sol
        qar_sp.write(f"{tag} ; psel_rand_src.lati", "", psel_rand_src.save_str(), skip_if_exist=True)
        qar_sp.write(f"{tag} ; psel_rand_src ; fu1.lat", "", sp_fu1.save_str(), skip_if_exist=True)
        qar_sp.write(f"{tag} ; psel_rand_src ; prop.lat", "", ps_prop_prs.save_str(), skip_if_exist=True)
        qar_sp.write(f"{tag}.lat", "", ps_prop.save_str(), skip_if_exist=True)
        qar_sp.flush()
        s_prop.save_float_from_double(sfw, tag, skip_if_exist=True)
        sfw.flush()
    for idx_psel in range(len(psel_list)):
        r = rs_ama.split(str(idx_psel)).u_rand_gen()
        assert 0 <= r and r <= 1
        for is_dagger in [ 0, 1, ]:
            inv_acc = 0
            compute_and_save(idx_psel, is_dagger, inv_acc)
            if r <= prob_acc_1_rand_sparse_u1:
                inv_acc = 1
                compute_and_save(idx_psel, is_dagger, inv_acc)
            if r <= prob_acc_2_rand_sparse_u1:
                inv_acc = 2
                compute_and_save(idx_psel, is_dagger, inv_acc)
    sfw.close()
    qar_sp.write("checkpoint.txt", "", "", skip_if_exist=True)
    qar_sp.flush()
    qar_sp.close()
    q.qrename_info(get_save_path(path_s + ".acc"), get_save_path(path_s))
    q.clean_cache(q.cache_inv)
    if is_test():
        q.json_results_append(f"{fname} {job_tag} {traj} inv_type={inv_type} psel_list_type={psel_list_type}")
        sfr = q.open_fields(get_load_path(path_s), "r")
        sfr_list = sfr.list()
        q.json_results_append(f"{fname} sfr.list()={sfr_list}")
        for tag in sfr_list:
            s_prop = q.SelProp()
            s_prop.load_double_from_float(sfr, tag)
            q.json_results_append(f"{fname} {tag}", q.get_data_sig_arr(s_prop, q.RngState(), 2), 1e-4)
        sfr.close()
        qar_sp = q.open_qar_info(get_load_path(path_sp + ".qar"), "r")
        qar_sp_list = qar_sp.list()
        qar_sp_list = q.get_comm().bcast(qar_sp_list)
        q.json_results_append(f"{fname} qar_sp.list()={qar_sp_list}")
        for tag in qar_sp_list:
            if tag.endswith(".lati"):
                ld = q.load_lat_data_int(get_load_path(f"{path_sp}/{tag}"))
            elif tag.endswith(".lat"):
                ld = q.load_lat_data(get_load_path(f"{path_sp}/{tag}"))
            elif tag == "checkpoint.txt":
                continue
            else:
                assert False
            q.json_results_append(f"{fname} {tag}", q.get_data_sig_arr(ld, q.RngState(), 2), 1e-4)
        qar_sp.close()
    q.release_lock()
    return [ f"{fname} {job_tag} {traj} {inv_type} done", ]

# -----------------------------------------------------------------------------

@q.timer_verbose
def compute_prop_rand_u1_type_acc(*, sfw, job_tag, traj, gf, eig, fsel, idx_rand_u1, inv_type, inv_acc):
    """
    same rand source for different inv_type
    """
    tag = f"idx_rand_u1={idx_rand_u1} ; type={inv_type} ; accuracy={inv_acc}"
    if tag in sfw:
        return
    q.check_stop()
    q.check_time_limit()
    inv = ru.get_inv(gf, job_tag, inv_type, inv_acc, eig=eig)
    seed = get_job_seed(job_tag)
    rs = q.RngState(f"seed {seed} {traj}").split(f"compute_prop_rand_u1(rand_u1)").split(str(idx_rand_u1))
    s_prop = q.mk_rand_u1_prop(inv, fsel, rs)
    s_prop.save_float_from_double(sfw, tag, skip_if_exist=True)
    sfw.flush()
    return s_prop

@q.timer_verbose
def compute_prop_rand_u1(*, job_tag, traj, inv_type, gf, path_s, fsel, eig=None):
    """
    use fsel instead of fselc
    """
    n_rand_u1_fsel = get_param(job_tag, "n_rand_u1_fsel")
    total_site = q.Coordinate(get_param(job_tag, "total_site"))
    geo = q.Geometry(total_site)
    sfw = q.open_fields(get_save_path(path_s + ".acc"), "a", q.Coordinate([ 2, 2, 2, 4, ]))
    def comp(idx_rand_u1, inv_acc):
        compute_prop_rand_u1_type_acc(
                sfw=sfw,
                job_tag=job_tag, traj=traj,
                gf=gf, eig=eig, fsel=fsel,
                idx_rand_u1=idx_rand_u1,
                inv_type=inv_type, inv_acc=inv_acc)
    seed = get_job_seed(job_tag)
    rs = q.RngState(f"seed {seed} {traj}").split(f"compute_prop_rand_u1(ama)")
    prob1 = get_param(job_tag, "prob_acc_1_rand_u1")
    prob2 = get_param(job_tag, "prob_acc_2_rand_u1")
    for idx_rand_u1 in range(n_rand_u1_fsel):
        r = rs.split(str(idx_rand_u1)).u_rand_gen()
        inv_acc = 0
        assert 0 <= r and r <= 1
        comp(idx_rand_u1, inv_acc)
        inv_acc = 1
        if r <= prob1:
            comp(idx_rand_u1, inv_acc)
        inv_acc = 2
        if r <= prob2:
            comp(idx_rand_u1, inv_acc)
    sfw.close()
    q.qrename_info(get_save_path(path_s + ".acc"), get_save_path(path_s))

@q.timer(is_timer_fork=True)
def run_prop_rand_u1(job_tag, traj, *, inv_type, get_gf, get_fsel, get_eig=None):
    fname = q.get_fname()
    if None in [ get_gf, get_fsel, ]:
        return
    if get_eig is None:
        if inv_type == 0:
            return
        get_eig = lambda: None
    inv_type_name_list = get_param(job_tag, "quark_flavor_list", default=[ "light", "strange", "charm", ])
    inv_type_name = inv_type_name_list[inv_type]
    path_s = f"{job_tag}/prop-rand-u1-{inv_type_name}/traj-{traj}"
    if get_load_path(path_s + "/geon-info.txt") is not None:
        return
    if q.obtain_lock(f"locks/{job_tag}-{traj}-{fname}-{inv_type_name}"):
        gf = get_gf()
        fsel = get_fsel()
        eig = get_eig()
        compute_prop_rand_u1(
                job_tag=job_tag, traj=traj,
                inv_type=inv_type,
                gf=gf,
                path_s=path_s,
                fsel=fsel,
                eig=eig)
        q.release_lock()
        return [ f"{fname} {job_tag} {traj} {inv_type} done", ]

# -----------------------------------------------------------------------------

@q.timer_verbose
def compute_prop_3(inv, src_smear, *, tag, sfw, qar_sp, sfw_m, psel, fsel, gt, psel_smear, fsel_smear_median, smear):
    sol = inv * src_smear
    sp_sol = q.PselProp(psel)
    sp_sol @= sol
    qar_sp.write(f"{tag}.lat", "", sp_sol.save_str(), skip_if_exist=True)
    sol_gt = gt * sol
    sol_ws = sol_gt.glb_sum_tslice()
    qar_sp.write(f"{tag} ; wsnk.lat", "", sol_ws.save_str(), skip_if_exist=True)
    sol_smear = smear(sol)
    sol_smear_psel = q.PselProp(psel_smear)
    sol_smear_psel @= sol_smear
    qar_sp.write(f"{tag} ; smear-snk.lat", "", sol_smear_psel.save_str(), skip_if_exist=True)
    sm_sol = q.SelProp(fsel_smear_median)
    sm_sol @= sol_smear
    sm_sol.save_float_from_double(sfw_m, f"{tag} ; smear-snk", skip_if_exist=True)
    s_sol = q.SelProp(fsel)
    s_sol @= sol
    qar_sp.flush()
    sfw_m.flush()
    s_sol.save_float_from_double(sfw, tag, skip_if_exist=True)
    sfw.flush()
    return sol

@q.timer
def compute_prop_smear(job_tag, xg_src, inv_type, inv_acc, *,
        idx, gf, gt, sfw, qar_sp, sfw_m, psel, fsel, psel_smear, fsel_smear_median, gf_ape, eig):
    xg = xg_src
    xg_str = f"({xg[0]},{xg[1]},{xg[2]},{xg[3]})"
    tag = f"smear ; xg={xg_str} ; type={inv_type} ; accuracy={inv_acc}"
    if tag in sfw:
        if 0 == q.get_id_node():
            assert f"{tag}.lat" in qar_sp
            assert f"{tag} ; wsnk.lat" in qar_sp
            assert f"{tag} ; smear-snk.lat" in qar_sp
        assert f"{tag} ; smear-snk" in sfw_m
        return None
    q.check_stop()
    q.check_time_limit()
    q.displayln_info(f"compute_prop_smear: {job_tag} idx={idx} tag='{tag}'")
    inv = ru.get_inv(gf, job_tag, inv_type, inv_acc, eig=eig)
    total_site = q.Coordinate(get_param(job_tag, "total_site"))
    geo = q.Geometry(total_site)
    coef = get_param(job_tag, "prop_smear_coef")
    step = get_param(job_tag, "prop_smear_step")
    def smear(src):
        return q.prop_smear(src, gf_ape, coef, step)
    src = smear(q.mk_point_src(geo, xg_src))
    prop = compute_prop_3(
            inv, src, tag=tag,
            sfw=sfw, qar_sp=qar_sp, sfw_m=sfw_m,
            psel=psel, fsel=fsel, gt=gt,
            psel_smear=psel_smear, fsel_smear_median=fsel_smear_median,
            smear=smear,
            )

@q.timer_verbose
def compute_prop_smear_all(
        job_tag, traj, *,
        inv_type, gf, gt, gf_ape, eig,
        psel, fsel, psel_smear, fsel_smear_median,
        ):
    inv_type_name_list = [ "light", "strange", ]
    inv_type_name = inv_type_name_list[inv_type]
    path_s = f"{job_tag}/prop-smear-{inv_type_name}/traj-{traj}"
    path_sm = f"{job_tag}/psel_smear_median-prop-smear-{inv_type_name}/traj-{traj}"
    path_sp = f"{job_tag}/psel-prop-smear-{inv_type_name}/traj-{traj}"
    sfw = q.open_fields(get_save_path(path_s + ".acc"), "a", q.Coordinate([ 2, 2, 2, 4, ]))
    sfw_m = q.open_fields(get_save_path(path_sm + ".acc"), "a", q.Coordinate([ 2, 2, 2, 4, ]))
    qar_sp = q.open_qar_info(get_save_path(path_sp + ".qar"), "a")
    def comp(idx, xg_src, inv_acc):
        compute_prop_smear(
                job_tag, xg_src, inv_type, inv_acc,
                idx=idx,
                gf=gf, gt=gt, gf_ape=gf_ape, eig=eig,
                sfw=sfw, qar_sp=qar_sp, sfw_m=sfw_m,
                psel=psel, fsel=fsel,
                psel_smear=psel_smear,
                fsel_smear_median=fsel_smear_median,
                )
    prob1 = get_param(job_tag, "prob_acc_1_smear")
    prob2 = get_param(job_tag, "prob_acc_2_smear")
    seed = get_job_seed(job_tag)
    rs = q.RngState(f"seed {seed} {traj}").split(f"compute_prop_smear_all(ama)")
    for idx, xg_src in enumerate(psel_smear):
        r = rs.split(f"{xg_src.to_tuple()}").u_rand_gen()
        assert 0 <= r and r <= 1
        comp(idx, xg_src, inv_acc=0)
        if r <= prob1:
            comp(idx, xg_src, inv_acc=1)
        if r <= prob2:
            comp(idx, xg_src, inv_acc=2)
    sfw.close()
    qar_sp.write("checkpoint.txt", "", "", skip_if_exist=True)
    qar_sp.flush()
    qar_sp.close()
    q.qrename_info(get_save_path(path_sm + ".acc"), get_save_path(path_sm))
    q.qrename_info(get_save_path(path_s + ".acc"), get_save_path(path_s))

@q.timer(is_timer_fork=True)
def run_prop_smear(job_tag, traj, *, inv_type, get_gf, get_gf_ape, get_eig, get_gt, get_psel, get_fsel, get_psel_smear, get_psel_smear_median):
    fname = q.get_fname()
    if None in [ get_gf, get_gt, get_gf_ape, get_psel, get_fsel, get_psel_smear, get_psel_smear_median, ]:
        return
    if get_eig is None:
        if inv_type == 0:
            return
        get_eig = lambda: None
    inv_type_name_list = [ "light", "strange", ]
    inv_type_name = inv_type_name_list[inv_type]
    if get_load_path(f"{job_tag}/prop-smear-{inv_type_name}/traj-{traj}/geon-info.txt") is not None:
        assert get_load_path(f"{job_tag}/psel-prop-smear-{inv_type_name}/traj-{traj}/checkpoint.txt") is not None
        assert get_load_path(f"{job_tag}/psel_smear_median-prop-smear-{inv_type_name}/traj-{traj}/geon-info.txt") is not None
        return
    if q.obtain_lock(f"locks/{job_tag}-{traj}-{fname}-{inv_type_name}"):
        gf = get_gf()
        gt = get_gt()
        eig = get_eig()
        psel = get_psel()
        fsel = get_fsel()
        assert fsel.is_containing(psel)
        psel_smear = get_psel_smear()
        psel_smear_median = get_psel_smear_median()
        fsel_smear_median = q.FieldSelection(psel_smear_median)
        assert psel_smear_median.is_containing(psel_smear)
        gf_ape = get_gf_ape()
        compute_prop_smear_all(
                job_tag, traj,
                inv_type=inv_type, gf=gf, gt=gt, gf_ape=gf_ape, eig=eig,
                psel=psel, fsel=fsel, psel_smear=psel_smear, fsel_smear_median=fsel_smear_median,
                )
        q.release_lock()
        return [ f"{fname} {job_tag} {traj} {inv_type} done", ]

# -----------------------------------------------------------------------------

# @q.timer_verbose
# def compute_prop_1(inv, src, *, tag, sfw, path_sp, psel, fsel):
#     fn_sp = os.path.join(path_sp, f"{tag}.lat")
#     fn_spw = os.path.join(path_sp, f"{tag} ; wsnk.lat")
#     sol = inv * src
#     sp_sol = q.PselProp(psel)
#     sp_sol @= sol
#     sp_sol.save(get_save_path(fn_sp))
#     sol_ws = sol.glb_sum_tslice()
#     sol_ws.save(get_save_path(fn_spw))
#     s_sol = q.SelProp(fsel)
#     s_sol @= sol
#     s_sol.save_float_from_double(sfw, tag)
#     sfw.flush()
#     return sol
#
# @q.timer
# def compute_prop_wsrc(gf, gt, tslice, job_tag, inv_type, inv_acc, *,
#         idx, sfw, path_sp, psel, fsel, eig, finished_tags):
#     tag = f"tslice={tslice} ; type={inv_type} ; accuracy={inv_acc}"
#     if tag in finished_tags:
#         return None
#     q.check_stop()
#     q.check_time_limit()
#     q.displayln_info(f"compute_prop_wsrc: idx={idx} tslice={tslice}", job_tag, inv_type, inv_acc)
#     inv = ru.get_inv(gf, job_tag, inv_type, inv_acc, gt=gt, eig=eig)
#     total_site = q.Coordinate(get_param(job_tag, "total_site"))
#     geo = q.Geometry(total_site)
#     src = q.mk_wall_src(geo, tslice)
#     prop = compute_prop_1(inv, src, tag=tag, sfw=sfw, path_sp=path_sp,
#                           psel=psel, fsel=fsel)
#
# @q.timer_verbose
# def compute_prop_wsrc_all(job_tag, traj, *,
#                           inv_type, gf, gt, wi, psel, fsel, eig):
#     inv_type_names = [ "light", "strange", ]
#     inv_type_name = inv_type_names[inv_type]
#     path_s = f"{job_tag}/prop-wsrc-{inv_type_name}/traj-{traj}"
#     path_sp = f"{job_tag}/psel-prop-wsrc-{inv_type_name}/traj-{traj}"
#     finished_tags = q.properly_truncate_fields(get_save_path(path_s + ".acc"))
#     sfw = q.open_fields(get_save_path(path_s + ".acc"), "a", q.Coordinate([ 2, 2, 2, 4, ]))
#     for inv_acc in [ 2, 1, ]:
#         for p in wi:
#             idx, tslice, inv_type_p, inv_acc_p=p
#             if inv_type_p == inv_type and inv_acc_p == inv_acc:
#                 compute_prop_wsrc(gf, gt, tslice, job_tag, inv_type, inv_acc,
#                         idx=idx, sfw=sfw, path_sp=path_sp,
#                         psel=psel, fsel=fsel, eig=eig,
#                         finished_tags=finished_tags)
#     sfw.close()
#     q.qtouch_info(get_save_path(os.path.join(path_sp, "checkpoint.txt")))
#     # q.qtouch_info(get_save_path(os.path.join(path_sp, "checkpoint ; wsnk.txt")))
#     q.qrename_info(get_save_path(path_s + ".acc"), get_save_path(path_s))
#     q.qar_create_info(get_save_path(path_sp + ".qar"), get_save_path(path_sp), is_remove_folder_after=True)
#     # q.qar_create_info(get_save_path(path_s + ".qar"), get_save_path(path_s), is_remove_folder_after=True)
#
# @q.timer
# def run_prop_wsrc(job_tag, traj, *, inv_type, get_gf, get_eig, get_gt, get_psel, get_fsel, get_wi):
#     """
#     Can use `run_prop_wsrc_sparse` instead.
#     #
#     run_prop_wsrc_full(job_tag, traj, inv_type=0, get_gf=get_gf, get_eig=get_eig, get_gt=get_gt, get_wi=get_wi)
#     get_fsel, get_psel, get_fsel_prob, get_psel_prob = run_fsel_psel_from_wsrc_prop_full(job_tag, traj, get_wi=get_wi)
#     run_prop_wsrc_sparse(job_tag, traj, inv_type=0, get_gt=get_gt, get_psel=get_psel, get_fsel=get_fsel, get_wi=get_wi)
#     """
#     if None in [ get_gf, get_gt, get_psel, get_fsel, ]:
#         return
#     if get_eig is None:
#         if inv_type == 0:
#             return
#         get_eig = lambda: None
#     inv_type_names = [ "light", "strange", ]
#     inv_type_name = inv_type_names[inv_type]
#     if get_load_path(f"{job_tag}/prop-wsrc-{inv_type_name}/traj-{traj}/geon-info.txt") is not None:
#         return
#     if q.obtain_lock(f"locks/{job_tag}-{traj}-wsrc-{inv_type_name}"):
#         gf = get_gf()
#         gt = get_gt()
#         eig = get_eig()
#         psel = get_psel()
#         fsel = get_fsel()
#         assert fsel.is_containing(psel)
#         wi = get_wi()
#         compute_prop_wsrc_all(job_tag, traj,
#                               inv_type=inv_type, gf=gf, gt=gt, wi=wi,
#                               psel=psel, fsel=fsel, eig=eig)
#         q.release_lock()



# all cpp level import

from qlat_utils.everything cimport *

cdef extern from "qlat/mpi.h" namespace "qlat":

    Long& mpi_level_count() except +
    Int begin(const Int id_node, const Coordinate& size_node, const Int color) except +
    Int end(const bool is_preserving_cache) except +
    const Coordinate& get_size_node() except +
    const Coordinate& get_coor_node() except +
    Int bcast(Long& x, const Int root) except +
    Int bcast(RealD& x, const Int root) except +
    Int bcast(RealF& x, const Int root) except +
    Int bcast(ComplexD& x, const Int root) except +
    Int bcast(ComplexF& x, const Int root) except +
    Int bcast(std_string& recv, const Int root) except +
    Int bcast(Coordinate& x, const Int root) except +
    Int bcast(LatData& ld, const Int root) except +
    Int bcast_any(Long& x, const bool b) except +
    Int bcast_any(RealD& x, const bool b) except +
    Int bcast_any(RealF& x, const bool b) except +
    Int bcast_any(ComplexD& x, const bool b) except +
    Int bcast_any(ComplexF& x, const bool b) except +
    Long f_bcast_any(const Long& x, const bool b) except +
    RealD f_bcast_any(const RealD& x, const bool b) except +
    RealF f_bcast_any(const RealF& x, const bool b) except +
    ComplexD f_bcast_any(const ComplexD& x, const bool b) except +
    ComplexF f_bcast_any(const ComplexF& x, const bool b) except +
    std_vector[Int]& get_id_node_list_for_shuffle() except +
    std_vector[Int]& get_id_node_in_shuffle_list() except +
    Int glb_sum(Long& x) except +
    Int glb_sum(RealD& x) except +
    Int glb_sum(RealF& x) except +
    Int glb_sum(ComplexD& x) except +
    Int glb_sum(ComplexF& x) except +
    Int glb_sum(LatData& ld) except +
    bool glb_all(const bool b) except +
    bool glb_any(const bool b) except +
    Long f_glb_sum(const Long& ld) except +
    RealD f_glb_sum(const RealD& ld) except +
    RealF f_glb_sum(const RealF& ld) except +
    ComplexD f_glb_sum(const ComplexD& ld) except +
    ComplexF f_glb_sum(const ComplexF& ld) except +

cdef extern from "qlat/geometry.h" namespace "qlat":

    cdef cppclass GeometryNode:
        bool initialized
        Int num_node
        Int id_node
        Coordinate size_node
        Coordinate coor_node
        GeometryNode()
        void init()
        void init(const Int id_node, const Coordinate& size_node) except +
        void init(const Coordinate& coor_node, const Coordinate& size_node) except +
    cdef cppclass Geometry:
        bool initialized
        GeometryNode geon
        Int eo
        Coordinate node_site
        Coordinate expansion_left
        Coordinate expansion_right
        Coordinate node_site_expanded
        bool is_only_local
        Geometry()
        void init()
        void init(const Coordinate& total_site) except +
        void init(const Coordinate& coor_node, const Coordinate& size_node, const Coordinate& node_site) except +
        void init(const Int& id_node, const Coordinate& size_node, const Coordinate& node_site) except +
        void reset_node_site_expanded() except +
        Coordinate total_site()
        Coordinate local_site()
        Long local_volume() except +
        Long local_volume_expanded() except +
        Long total_volume() except +
        Long spatial_volume() except +
        bool is_local(const Coordinate& x)
        Coordinate coordinate_g_from_l(const Coordinate& xl)
        Coordinate coordinate_l_from_g(const Coordinate& xg)
        Long index_from_coordinate(const Coordinate& xl)
        Coordinate coordinate_from_index(const Long index)
        Long index_from_g_coordinate(const Coordinate& xg)
        Coordinate g_coordinate_from_index(const Long index)
    bool operator==(const Geometry& geo1, const Geometry& geo2) except +
    bool operator!=(const Geometry& geo1, const Geometry& geo2) except +
    std_string show(const Geometry& geo) except +
    Geometry geo_resize(const Geometry& geo, Int thick) except +
    Geometry geo_resize(const Geometry& geo, const Coordinate& expansion_left, const Coordinate& expansion_right) except +
    Geometry geo_eo(const Geometry& geo, Int eo) except +

cdef extern from "qlat/utils-io.h" namespace "qlat":

    void release_lock() except +
    bool obtain_lock(const std_string& path) except +
    void qquit(const std_string& msg) except +
    void check_time_limit(const RealD budget) except +
    void check_stop(const std_string& fn) except +

cdef extern from "qlat/core.h" namespace "qlat":

    cdef cppclass PointsDistType:
        pass
    std_string show(const PointsDistType points_dist_type) except +
    PointsDistType read_points_dist_type(const std_string& points_dist_type_str) except +
    cdef cppclass PointsSelection:
        bool initialized
        PointsDistType points_dist_type
        Coordinate total_site
        PointsSelection()
        PointsSelection(const Coordinate& total_site, const Long n_points) except +
        PointsSelection(const Coordinate& total_site, const Long n_points, const PointsDistType points_dist_type) except +
        void init()
        void init(const Coordinate& total_site, const Long n_points) except +
        void init(const Coordinate& total_site, const Long n_points, const PointsDistType points_dist_type) except +
        Long size()
        Coordinate* data()
        Coordinate& operator[](Long i) except +
    bool operator==(const PointsSelection& psel1, const PointsSelection& psel2) except +
    bool operator!=(const PointsSelection& psel1, const PointsSelection& psel2) except +
    void qswap(PointsSelection& x, PointsSelection& y) except +
    cdef cppclass SelectedPoints[T]:
        bool initialized;
        PointsDistType points_dist_type
        Int multiplicity
        Long n_points
        vector[T] points
        SelectedPoints()
        void init()
        void init(const Long n_points, const Int multiplicity, const PointsDistType points_dist_type) except +
        void init(const PointsSelection& psel, const Int multiplicity) except +
    Int bcast(PointsSelection& psel, const Int root) except +
    Int bcast[T](SelectedPoints[T]& sp, const Int root) except +
    cdef cppclass Field[T]:
        Int multiplicity
        vector[T] field
        Field()
        void init()
        void init(const Geometry& geo) except +
        void init(const Geometry& geo, Int multiplicity) except +
        void init(const Field[T]& field) except +
        Geometry get_geo() except +
        T& get_elem(const Coordinate& x) except +
        T& get_elem(const Coordinate& x, const Int m) except +
        T& get_elem(const Long index) except +
        T& get_elem(const Long index, const Int m) except +
        Vector[T] get_elems(const Coordinate& x) except +
        Vector[T] get_elems(const Long index) except +
    cdef cppclass GaugeField(Field[ColorMatrix]):
        pass
    cdef cppclass GaugeMomentum(Field[ColorMatrix]):
        pass
    cdef cppclass GaugeTransform(Field[ColorMatrix]):
        pass
    cdef cppclass Prop(Field[WilsonMatrix]):
        pass
    cdef cppclass FermionField4d(Field[WilsonVector]):
        pass
    cdef cppclass FieldRank(Field[Int64t]):
        pass
    cdef cppclass FieldIndex(Field[Long]):
        pass
    cdef cppclass FieldSelection:
        Long n_elems
        FieldRank f_rank
        FieldIndex f_local_idx
        vector[Int64t] ranks
        vector[Long] indices
        FieldSelection()
        void init()
        Geometry get_geo() except +
    void set_psel_from_fsel(PointsSelection& psel, const FieldSelection& fsel) except +
    void set_fsel_from_psel(FieldSelection& fsel, const PointsSelection& psel, const Geometry& geo, const Long rank_psel) except +
    void set_geo_from_psel(Geometry& geo, const PointsSelection& psel) except +
    cdef cppclass SelectedField[T]:
        Long n_elems;
        Int multiplicity
        vector[T] field
        SelectedField()
        void init()
        void init(const Geometry& geo, const Long n_elems, const Int multiplicity) except +
        void init(const FieldSelection& fsel, const Int multiplicity) except +
        Geometry get_geo() except +
    Vector[T] get_data[T](const Field[T]& x) except +
    void set_zero[T](Field[T]& x) except +
    void set_unit[T](Field[T]& f, const ComplexD& coef) except +
    void qswap[T](Field[T]& x, Field[T]& y) except +
    Vector[T] get_data[T](const SelectedField[T]& x) except +
    void set_zero[T](SelectedField[T]& x) except +
    void qswap[T](SelectedField[T]& x, SelectedField[T]& y) except +
    Vector[T] get_data[T](const SelectedPoints[T]& x) except +
    void set_zero[T](SelectedPoints[T]& x) except +
    void qswap[T](SelectedPoints[T]& x, SelectedPoints[T]& y) except +
    void qswap_cast[M](PointsSelection& f1, SelectedPoints[M]& f2, Coordinate& total_site2) except +
    void qswap_cast[M,N](Field[M]& f1, Field[N]& f2) except +
    void qswap_cast[M,N](SelectedField[M]& f1, SelectedField[N]& f2) except +
    void qswap_cast[M,N](Field[M]& f1, SelectedPoints[N]& f2, Geometry& geo2) except +
    void qswap_cast[M,N](SelectedField[M]& f1, SelectedPoints[N]& f2, Geometry& geo2) except +
    void qswap_cast[M,N](SelectedPoints[M]& f1, SelectedPoints[N]& f2) except +
    cdef cppclass SelProp(SelectedField[WilsonMatrix]):
        pass
    cdef cppclass PselProp(SelectedPoints[WilsonMatrix]):
        pass
    void set_field_m(Field[Char]& f, const Field[Char]& f1, const Int m, const Int m1, const Int sizeof_m) except +

cdef extern from "qlat/field.h" namespace "qlat":

    RealD qnorm[M](const Field[M]& f) except +
    void qnorm_field[M](Field[RealD]& f, const Field[M]& f1) except +
    void set_u_rand[M](Field[M]& sp, const RngState& rs, const RealD upper, const RealD lower) except +
    void set_g_rand[M](Field[M]& sp, const RngState& rs, const RealD center, const RealD sigma) except +
    void set_xg_field(Field[Int]& f, const Geometry& geo) except +
    void field_shift[M](Field[M]& f, const Field[M]& f1, const Coordinate& shift) except +
    void reflect_field[M](Field[M]& f) except +
    void set_sqrt_field(Field[RealD]& f, const Field[RealD]& f1) except +

cdef extern from "qlat/field-expand.h" namespace "qlat":

    cdef cppclass CommPlan:
        CommPlan()

cdef extern from "qlat/field-io.h" namespace "qlat":

    bool is_field(const std_string& path) except +
    bool is_d_field(const std_string& path) except +
    bool dist_repartition(const Coordinate& new_size_node, const std_string& path, const std_string& new_path) except +
    crc32_t field_crc32[M](const Field[M]& f) except +
    Long write_field[M](const Field[M]& f, const std_string& path, const Coordinate& new_size_node) except +
    Long read_field[M](Field[M]& f, const std_string& path, const Coordinate& new_size_node) except +

cdef extern from "qlat/selected-field-io.h" namespace "qlat":

    Long write_field_selection(const FieldSelection& fsel, const std_string& path) except +
    Long read_field_selection(FieldSelection& fsel, const std_string& path) except +
    #
    void mk_field_selection(FieldRank& f_rank, const Geometry& geo, const Int64t val) except +
    void mk_field_selection(FieldRank& f_rank, const Coordinate& total_site, const Long n_per_tslice, const RngState& rs) except +
    void add_field_selection(FieldRank& f_rank, const PointsSelection& psel, const Long rank_psel) except +
    void add_field_selection(FieldRank& f_rank, const FieldSelection& fsel) except +
    void update_field_selection(FieldSelection& fsel) except +
    #
    PointsSelection psel_from_fsel(const FieldSelection& fsel) except +
    PointsSelection psel_from_fsel_local(const FieldSelection& fsel) except +
    bool is_matching_fsel(const FieldSelection& fsel1, const FieldSelection& fsel2) except +
    bool is_containing(const FieldSelection& fsel, const FieldSelection& fsel_small) except +
    bool is_containing(const FieldSelection& fsel, const PointsSelection& psel_small) except +
    bool is_containing(const PointsSelection& psel, const FieldSelection& fsel_small) except +
    bool is_containing(const PointsSelection& psel, const PointsSelection& psel_small) except +
    void intersect_with(FieldSelection& fsel, const FieldSelection& fsel1) except +
    PointsSelection intersect(const FieldSelection& fsel, const PointsSelection& psel) except +

cdef extern from "qlat/selected-shuffle.h" namespace "qlat":

    cdef cppclass SelectedShufflePlan:
        PointsDistType points_dist_type_send
        PointsDistType points_dist_type_recv
        Coordinate total_site
        vector[Coordinate] size_node_send
        vector[Coordinate] coor_node_send
        vector[Coordinate] size_node_recv
        vector[Coordinate] coor_node_recv
        Long num_selected_points_send
        Long num_selected_points_recv
        vector[Long] n_points_selected_points_send
        vector[Long] n_points_selected_points_recv
        SelectedPoints[Long] shuffle_idx_points_send
        SelectedPoints[Long] shuffle_idx_points_recv
        SelectedPoints[Long] shuffle_idx_points_local
        Long total_count_send
        Long total_count_recv
        Long total_count_local
        vector[Long] sendcounts
        vector[Long] recvcounts
        vector[Long] sdispls
        vector[Long] rdispls
        void init() except +
    #
    void set_selected_shuffle_plan_l_from_g(SelectedShufflePlan& ssp, const PointsSelection& psel, const Int root) except +
    void set_selected_shuffle_plan_l_from_g(SelectedShufflePlan& ssp, const std_vector[PointsSelection]& psel_vec, const std_vector[Int]& root_vec) except +
    void set_selected_shuffle_plan_g_from_l(SelectedShufflePlan& ssp, const PointsSelection& psel, const Int root, const Geometry& geo) except +
    void set_selected_shuffle_plan_g_from_l(SelectedShufflePlan& ssp, const std_vector[PointsSelection]& psel_vec, const std_vector[Int]& root_vec, const std_vector[Geometry]& geo_vec) except +
    void set_selected_shuffle_plan_r_from_l(SelectedShufflePlan& ssp, const PointsSelection& psel, const Geometry& geo, const RngState& rs) except +
    void set_selected_shuffle_plan_r_from_l(SelectedShufflePlan& ssp, const std_vector[PointsSelection]& psel_vec, const std_vector[Geometry]& geo_vec, const RngState& rs) except +
    void set_selected_shuffle_plan_dist_r_from_l(SelectedShufflePlan& ssp, const PointsSelection& psel, const Geometry& geo, const RngState& rs, const std_vector[Int]& id_node_vec) except +
    void set_selected_shuffle_plan_t_slice_from_l(SelectedShufflePlan& ssp, const std_vector[PointsSelection]& psel_vec, const std_vector[Geometry]& geo_vec) except +
    void set_selected_shuffle_plan_dist_t_slice_from_l(SelectedShufflePlan& ssp, const PointsSelection& psel, const Geometry& geo, const Int num_field) except +
    #
    void shuffle_selected_points_char(SelectedPoints[Char]& spc, const SelectedPoints[Char]& spc0, const SelectedShufflePlan& ssp) except +
    void shuffle_selected_points_back_char(SelectedPoints[Char]& spc, const SelectedPoints[Char]& spc0, const SelectedShufflePlan& ssp) except +
    void shuffle_selected_points_char(std_vector[SelectedPoints[Char]]& spc_vec, const std_vector[SelectedPoints[Char]]& spc0_vec, const SelectedShufflePlan& ssp) except +
    void shuffle_selected_points_back_char(std_vector[SelectedPoints[Char]]& spc_vec, const std_vector[SelectedPoints[Char]]& spc0_vec, const SelectedShufflePlan& ssp) except +
    #
    void shuffle_points_selection(PointsSelection& psel, const PointsSelection& psel0, const SelectedShufflePlan& ssp) except +
    void shuffle_points_selection_back(PointsSelection& psel, const PointsSelection& psel0, const SelectedShufflePlan& ssp) except +
    void shuffle_points_selection(std_vector[PointsSelection]& psel_vec, const std_vector[PointsSelection]& psel0_vec, const SelectedShufflePlan& ssp) except +
    void shuffle_points_selection_back(std_vector[PointsSelection]& psel_vec, const std_vector[PointsSelection]& psel0_vec, const SelectedShufflePlan& ssp) except +
    #
    void shuffle_selected_points[M](SelectedPoints[M]& sp, const SelectedPoints[M]& sp0, const SelectedShufflePlan& ssp) except +
    void shuffle_selected_points_back[M](SelectedPoints[M]& sp, const SelectedPoints[M]& sp0, const SelectedShufflePlan& ssp) except +
    void shuffle_selected_points[M](std_vector[SelectedPoints[M]]& sp_vec, const std_vector[SelectedPoints[M]]& sp0_vec, const SelectedShufflePlan& ssp) except +
    void shuffle_selected_points_back[M](std_vector[SelectedPoints[M]]& sp_vec, const std_vector[SelectedPoints[M]]& sp0_vec, const SelectedShufflePlan& ssp) except +
    #
    void shuffle_field_selection(PointsSelection& psel, const FieldSelection& fsel0, const SelectedShufflePlan& ssp) except +
    void shuffle_selected_field[M](SelectedPoints[M]& sp, const SelectedField[M]& sf, const SelectedShufflePlan& ssp) except +

cdef extern from "qlat/selected-points.h" namespace "qlat":

    PointsSelection mk_tslice_points_selection(const Coordinate& total_site, const Int t_dir) except +
    PointsSelection mk_random_points_selection(const Coordinate& total_site, const Long num, const RngState& rs) except +
    void set_psel_full(PointsSelection& psel, const Geometry& geo) except +
    void lat_data_from_points_selection(LatDataInt& ld, const PointsSelection& psel) except +
    void points_selection_from_lat_data(PointsSelection& psel, const LatDataInt& ld) except +
    void points_selection_from_lat_data(PointsSelection& psel, const LatDataInt& ld, const PointsDistType points_dist_type) except +
    void save_points_selection(const PointsSelection& psel, const std_string& path) except +
    void save_points_selection_info(const PointsSelection& psel, const std_string& path) except +
    PointsSelection load_points_selection(const std_string& path) except +
    PointsSelection load_points_selection_info(const std_string& path) except +
    #
    RealD qnorm[M](const SelectedPoints[M]& sp) except +
    void qnorm_field[M](SelectedPoints[RealD]& sp, const SelectedPoints[M]& sp1) except +
    void set_u_rand[M](SelectedPoints[M]& sp, const PointsSelection& psel, const RngState& rs, const RealD upper, const RealD lower) except +
    void set_g_rand[M](SelectedPoints[M]& sp, const PointsSelection& psel, const RngState& rs, const RealD center, const RealD sigma) except +
    void lat_data_from_selected_points[M](LatData& ld, const SelectedPoints[M]& sp) except +
    void selected_points_from_lat_data[M](SelectedPoints[M]& sp, const LatData& ld) except +
    void lat_data_from_selected_points[M](LatDataRealF& ld, const SelectedPoints[M]& sp) except +
    void selected_points_from_lat_data[M](SelectedPoints[M]& sp, const LatDataRealF& ld) except +
    void lat_data_from_selected_points[M](LatDataLong& ld, const SelectedPoints[M]& sp) except +
    void selected_points_from_lat_data[M](SelectedPoints[M]& sp, const LatDataLong& ld) except +
    void lat_data_from_selected_points[M](LatDataInt& ld, const SelectedPoints[M]& sp) except +
    void selected_points_from_lat_data[M](SelectedPoints[M]& sp, const LatDataInt& ld) except +
    void save_selected_points[M](const SelectedPoints[M]& sp, QFile& qfile, const bool is_sync_node) except +
    void load_selected_points[M](SelectedPoints[M]& sp, QFile& qfile, const bool is_sync_node) except +
    void save_selected_points[M](const SelectedPoints[M]& sp, const std_string& fn, const bool is_sync_node) except +
    void load_selected_points[M](SelectedPoints[M]& sp, const std_string& fn, const bool is_sync_node) except +
    std_string save_selected_points_str[M](const SelectedPoints[M]& sp, const bool is_sync_node) except +
    void load_selected_points_str[M](SelectedPoints[M]& sp, std_string& content, const bool is_sync_node) except +
    void field_glb_sum[M](SelectedPoints[M]& sp, const Field[M]& f) except +
    void field_glb_sum_tslice[M](SelectedPoints[M]& sp, const Field[M]& f, const Int t_dir) except +
    void field_glb_max[M](SelectedPoints[M]& sp, const Field[M]& f) except +
    void field_glb_min[M](SelectedPoints[M]& sp, const Field[M]& f) except +
    void set_sqrt_field(SelectedPoints[RealD]& sp, const SelectedPoints[RealD]& sp1) except +

cdef extern from "qlat/selected-field-io.h" namespace "qlat":

    RealD qnorm[M](const SelectedField[M]& sf) except +
    void qnorm_field[M](SelectedField[RealD]& f, const SelectedField[M]& f1) except +
    void set_u_rand[M](SelectedField[M]& sp, const FieldSelection& fsel, const RngState& rs, const RealD upper, const RealD lower) except +
    void set_g_rand[M](SelectedField[M]& sp, const FieldSelection& fsel, const RngState& rs, const RealD center, const RealD sigma) except +
    void set_selected_field[t](SelectedField[t]& sf, const Field[t]& f,
                               const FieldSelection& fsel) except +
    void set_selected_field[t](SelectedField[t]& sf, const SelectedField[t] sf0,
                               const FieldSelection& fsel, const FieldSelection& fsel0) except +
    void set_selected_field[t](SelectedField[t]& sf, const SelectedPoints[t] sp,
                               const FieldSelection& fsel, const PointsSelection& psel) except +
    void set_selected_points[t](SelectedPoints[t]& sp, const Field[t] f,
                                const PointsSelection& psel) except +
    void set_selected_points[t](SelectedPoints[t]& sp, const SelectedField[t] sf,
                                const PointsSelection& psel, const FieldSelection& fsel) except +
    void set_selected_points[t](SelectedPoints[t]& sp, const SelectedPoints[t] sp0,
                                const PointsSelection& psel, const PointsSelection& psel0) except +
    void set_field_selected[t](Field[t]& f, const SelectedField[t]& sf,
                               const FieldSelection& fsel) except +
    void set_field_selected[t](Field[t]& f, const SelectedPoints[t]& sp,
                               const Geometry& geo, const PointsSelection& psel) except +
    void set_selected_points[t](SelectedPoints[t]& sp, const Field[t] f,
                                const PointsSelection& psel, const Int m) except +
    void set_field_selected[t](Field[t]& f, const SelectedPoints[t]& sp,
                               const Geometry& geo, const PointsSelection& psel,
                               const Int multiplicity, const Int m) except +
    void set_sqrt_field(SelectedField[RealD]& f, const SelectedField[RealD]& f1) except +
    #
    bool is_selected_field(const std_string& path) except +
    Long write_selected_field[M](const SelectedField[M]& sf, const std_string& path, const FieldSelection& fsel) except +
    Long read_selected_field[M](SelectedField[M]& sf, const std_string& path, const FieldSelection& fsel) except +

cdef extern from "qlat/field-shuffle.h" namespace "qlat":

    void field_shift[M](SelectedField[M]& sf, FieldSelection& fsel,
            const SelectedField[M]& sf0, const FieldSelection& fsel0,
            const Coordinate& shift, const bool is_reflect) except +
    void shuffle_field[M](std_vector[Field[M]]& fs, const Field[M]& f, const Coordinate& new_size_node) except +
    void shuffle_field_back[M](Field[M]& f, const std_vector[Field[M]]& fs, const Coordinate& new_size_node) except +

cdef extern from "qlat/gauge-action.h" namespace "qlat":

    cdef cppclass GaugeAction:
        bool initialized
        RealD beta
        RealD c1
        void init() except +
        GaugeAction() except +

cdef extern from "qlat/qcd-prop.h" namespace "qlat":

    void set_ff_vec_from_prop(std_vector[FermionField4d]& ff_vec, const Prop& prop) except +
    void set_prop_from_ff_vec(Prop& prop, const std_vector[FermionField4d]& ff_vec) except +
    void set_wall_src(Prop& prop, const Geometry& geo_input, const Int tslice, const CoordinateD& lmom) except +
    void set_point_src(Prop& prop, const Geometry& geo_input, const Coordinate& xg, const ComplexD& value) except +
    void set_rand_vol_u1(Field[ComplexD]& fu1, const Geometry& geo_input, const Int multiplicity, const RngState& rs) except +
    void set_rand_vol_u1_src(Prop& prop, const Field[ComplexD]& fu1) except +
    void free_invert(
        Prop& p_sol, const Prop& p_src,
        const RealD mass, const RealD m5,
        const CoordinateD& momtwist) except +

cdef extern from "qlat/qcd-smear.h" namespace "qlat":

    void gf_ape_smear(GaugeField& gf, const GaugeField& gf0, const RealD alpha, const Long steps) except +
    void gf_spatial_ape_smear(GaugeField& gf, const GaugeField& gf0, const RealD alpha, const Long steps) except +
    void gf_hyp_smear(GaugeField& gf, const GaugeField& gf0, const RealD alpha1, const RealD alpha2, const RealD alpha3) except +
    void prop_smear(Prop& prop, const GaugeField& gf1, const RealD coef, const Int step, const CoordinateD& mom, const bool smear_in_time_dir) except +
    void prop_spatial_smear_no_comm(std_vector[FermionField4d]& ff_vec, const GaugeField& gf, const RealD coef, const Long step, const CoordinateD& mom) except +
    void gf_reduce_half(GaugeField& hgf, const GaugeField& gf) except +
    void prop_smear_qlat_convension(Prop& prop, const GaugeField& gf1, const RealD coef, const Int step, const CoordinateD& mom, const bool smear_in_time_dir, const Int mode_smear) except +

cdef extern from "qlat/qcd.h" namespace "qlat":

    RealD gf_avg_plaq(const GaugeField& gf) except +
    RealD gf_avg_spatial_plaq(const GaugeField& gf) except +
    RealD gf_avg_link_trace(const GaugeField& gf) except +
    void gf_plaq_field(Field[RealD]& f_plaq, const GaugeField& gf) except +
    RealD gf_local_avg_plaq(const GaugeField& gf, const Coordinate& block_site) except +
    void unitarize(Field[ColorMatrix]& gf) except +
    void make_tr_less_anti_herm_matrix(Field[ColorMatrix]& fc) except +

cdef extern from "qlat/qcd-utils.h" namespace "qlat":

    void set_left_expanded_gauge_field(GaugeField& gf1, const GaugeField& gf) except +

cdef extern from "qlat/qcd-gauge-transformation.h" namespace "qlat":

    void gt_apply_gauge_transformation(GaugeTransform& gt0, const GaugeTransform& gt1) except +
    void gt_apply_gauge_transformation(GaugeTransform& gt, const GaugeTransform& gt0, const GaugeTransform& gt1) except +
    void gf_apply_gauge_transformation(GaugeField& gf, const GaugeField& gf0, const GaugeTransform& gt, const bool is_dagger) except +
    void gt_invert(GaugeTransform& gt, const GaugeTransform& gt0) except +
    void ff_apply_gauge_transformation(FermionField4d& ff, const FermionField4d& ff0, const GaugeTransform& gt) except +
    void prop_apply_gauge_transformation(Prop& prop, const Prop& prop0, const GaugeTransform& gt) except +
    void prop_apply_gauge_transformation(SelectedField[WilsonMatrix]& prop, const SelectedField[WilsonMatrix]& prop0, const GaugeTransform& gt, const FieldSelection& fsel) except +
    void prop_apply_gauge_transformation(SelectedPoints[WilsonMatrix]& prop, const SelectedPoints[WilsonMatrix]& prop0, const GaugeTransform& gt, const PointsSelection& psel) except +

cdef extern from "qlat/hmc.h" namespace "qlat":

    bool metropolis_accept(RealD& accept_prob, const RealD delta_h, const int traj, const RngState& rs_) except +
    void set_rand_gauge_momentum(GaugeMomentum& gm, const RealD sigma, const RngState& rs) except +
    void set_rand_gauge_momentum(GaugeMomentum& gm, const Field[RealD]& mf, const RngState& rs) except +
    RealD gm_hamilton_node(const GaugeMomentum& gm) except +
    RealD gm_hamilton_node(const GaugeMomentum& gm, const Field[RealD]& mf) except +
    RealD gf_hamilton_node(const GaugeField& gf, const GaugeAction& ga) except +
    RealD gf_hamilton(const GaugeField& gf, const GaugeAction& ga) except +
    void gf_evolve(GaugeField& gf, const GaugeMomentum& gm, const RealD step_size) except +
    void gf_evolve_dual(GaugeField& gf, const GaugeMomentum& gm_dual, const RealD step_size) except +
    void gf_evolve(GaugeField& gf, const GaugeMomentum& gm, const Field[RealD]& mf, const RealD step_size) except +
    void gf_evolve_dual(GaugeField& gf, const GaugeMomentum& gm_dual, const Field[RealD]& mf_dual, const RealD step_size) except +
    void field_color_matrix_exp(Field[ColorMatrix]& fc, const Field[ColorMatrix]& fc1, const ComplexD coef) except +
    void field_color_matrix_mul(Field[ColorMatrix]& fc, const Field[ColorMatrix]& fc1, const Field[ColorMatrix]& fc2) except +
    void set_gm_force(GaugeMomentum& gm_force, const GaugeField& gf, const GaugeAction& ga) except +
    void set_gm_force_dual(GaugeMomentum& gm_force_dual, const GaugeField& gf, const GaugeMomentum& gm_force) except +
    RealD project_gauge_transform(GaugeMomentum& gm, GaugeMomentum& gm_dual, const Field[RealD]& mf, const Field[RealD]& mf_dual) except +
    void set_gauge_transform_momentum(GaugeMomentum& gm, GaugeMomentum& gm_dual, const Field[ColorMatrix]& gtm) except +
    void dot_gauge_momentum(Field[RealD]& f, const GaugeMomentum& gm1, const GaugeMomentum& gm2) except +
    void set_anti_hermitian_matrix_from_basis(Field[ColorMatrix]& fc, const Field[RealD]& basis) except +
    void set_basis_from_anti_hermitian_matrix(Field[RealD]& basis, const Field[ColorMatrix]& fc) except +

cdef extern from "qlat/fields-io.h" namespace "qlat":

    cdef cppclass ShuffledBitSet:
        FieldSelection fsel
        std_vector[FieldSelection] fsels
        ShuffledBitSet()
    ShuffledBitSet mk_shuffled_bitset(const FieldRank& f_rank, const Coordinate& new_size_node) except +
    ShuffledBitSet mk_shuffled_bitset(const FieldSelection& fsel, const Coordinate& new_size_node) except +
    ShuffledBitSet mk_shuffled_bitset(const Coordinate& total_site, const PointsSelection& xgs, const Coordinate& new_size_node) except +
    ShuffledBitSet mk_shuffled_bitset(const FieldRank& f_rank, const PointsSelection& xgs, const Coordinate& new_size_node) except +
    cdef cppclass ShuffledFieldsWriter:
        std_string path
        Coordinate new_size_node
        ShuffledFieldsWriter()
        void init() except +
        void init(const std_string& path_, const Coordinate& new_size_node_, const bool is_append) except +
        void close() except +
    cdef cppclass ShuffledFieldsReader:
        std_string path
        Coordinate new_size_node
        ShuffledFieldsReader()
        void init() except +
        void init(const std_string& path_, const Coordinate& new_size_node_) except +
        void close() except +
    Long write[M](ShuffledFieldsWriter& sfw, const std_string& fn, const Field[M]& field) except +
    Long write[M](ShuffledFieldsWriter& sfw, const std_string& fn, const ShuffledBitSet& sbs, const SelectedField[M]& sf) except +
    Long read[M](ShuffledFieldsReader& sfr, const std_string& fn, Field[M]& field) except +
    Long read[M](ShuffledFieldsReader& sfr, const std_string& fn, SelectedField[M]& sf, FieldSelection& fsel) except +
    Long read[M](ShuffledFieldsReader& sfr, const std_string& fn, const ShuffledBitSet& sbs, SelectedField[M]& sf) except +
    Long flush(ShuffledFieldsWriter& sfw) except +
    void read_through_sync_node(ShuffledFieldsReader& sfr) except +
    bool does_file_exist_sync_node(const ShuffledFieldsReader& sfr, const std_string& fn) except +
    bool does_file_exist_sync_node(const ShuffledFieldsWriter& sfw, const std_string& fn) except +
    bool is_sparse_field_sync_node(const ShuffledFieldsReader& sfr, const std_string& fn) except +
    std_vector[std_string] list_fields(const ShuffledFieldsReader& sfr) except +
    std_vector[std_string] list_fields(const ShuffledFieldsWriter& sfw) except +
    Int truncate_fields_sync_node(const std_string& path, const std_vector[std_string]& fns_keep, const Coordinate& new_size_node) except +
    std_vector[std_string] properly_truncate_fields_sync_node(const std_string& path, const bool is_check_all, const bool is_only_check, const Coordinate& new_size_node) except +
    bool has_duplicates(const ShuffledFieldsReader& sfr) except +
    void fields_build_index(const ShuffledFieldsReader& sfr) except +
    void fields_build_index(const std_string& path) except +
    #
    std_vector[std_string] show_all_shuffled_fields_writer() except +

cdef extern from "qlat/compressed-eigen-io.h" namespace "qlat":

    bool check_compressed_eigen_vectors(const std_string& path) except +
    bool eigen_system_repartition(const Coordinate& new_size_node, const std_string& path, const std_string& new_path) except +

cdef extern from "qlat/wilson-flow.h" namespace "qlat":

    void set_wilson_flow_z(GaugeMomentum& z, const GaugeField& gf, const RealD c1) except +
    void gf_wilson_flow_step_euler(GaugeField& gf, const RealD epsilon, const RealD c1) except +
    void gf_wilson_flow_step(GaugeField& gf, const RealD epsilon, const RealD c1) except +
    void gf_energy_density_dir_field(Field[RealD]& fd, const GaugeField& gf) except +
    void gf_energy_density_field(Field[RealD]& fd, const GaugeField& gf) except +
    RealD gf_energy_density(const GaugeField& gf) except +
    void set_plaq_flow_z(GaugeMomentum& z, const GaugeField& gf, const Field[RealD]& plaq_factor) except +
    void gf_block_stout_smear(GaugeField& gf, const GaugeField& gf0, const Coordinate& block_site, const RealD step_size) except +
    void gf_local_stout_smear(GaugeField& gf, const GaugeField& gf0, const Coordinate& block_site, const RealD step_size) except +
    void set_local_tree_gauge_f_dir(Field[Int]& f_dir, const Geometry& geo, const Coordinate& block_site, const bool is_uniform, const RngState& rs) except +
    void gt_local_tree_gauge(GaugeTransform& gt_inv, const GaugeField& gf, const Field[Int]& f_dir, const Int num_step) except +

cdef extern from "qlat/qcd-topology.h" namespace "qlat":

    RealD topology_charge_5(const GaugeField& gf) except +
    void clf_plaq_action_density_field(Field[RealD]& topf, const GaugeField& gf) except +
    void clf_spatial_plaq_action_density_field(Field[RealD]& topf, const GaugeField& gf) except +
    void clf_topology_field(Field[RealD]& topf, const GaugeField& gf) except +
    void clf_topology_field_5(Field[RealD]& topf, const GaugeField& gf) except +
    void clf_topology_field_5_terms(Field[RealD]& topf, const GaugeField& gf) except +

cdef extern from "qlat/muon-line.h" namespace "qlat":

    bool has_cuba() except +
    #
    void test_integrationMultidimensional() except +
    #
    void clear_muon_line_interpolations() except +
    Long get_number_of_muon_line_interpolations() except +
    #
    cdef cppclass IntegrationEps:
        RealD epsabs
        RealD epsrel
        Long mineval
        Long maxeval
        IntegrationEps()
    # epsabs = 1e-8
    # epsrel = 1e-3
    # mineval = 1024 * 1024
    # maxeval = 1024 * 1024 * 1024
    # dims = [ 6, 6, 6, 6, 6, ] ~ [ 16, 16, 16, 16, 16, ]
    bool compute_save_muonline_interpolation_cc(
            const std_string& path,
            const std_vector[Int]& dims,
            const IntegrationEps& eps) except +
    std_vector[RealD] muon_line_sym_py(
            const CoordinateD& x,
            const CoordinateD& y,
            const IntegrationEps& eps) except +
    #
    bool load_multiple_muonline_interpolations(const std_string& path, std_vector[Long] idx_list) except +
    #
    std_vector[std_vector[RealD]] get_muon_line_m_extra_weights_default() except +
    std_vector[std_vector[RealD]]& get_muon_line_m_extra_weights() except +
    void set_muon_line_m_extra_weights(const std_vector[std_vector[RealD]]& weight) except +
    #
    std_vector[RealD] get_muon_line_m_py(
            const CoordinateD& x, const CoordinateD& y,
            const CoordinateD& z,
            const Int idx,
            const IntegrationEps& eps) except +
    std_vector[RealD] get_muon_line_m_extra_py(const CoordinateD& x,
            const CoordinateD& y,
            const CoordinateD& z,
            const Int tag) except +
    std_vector[RealD] get_muon_line_m_extra_lat_py(
            const Coordinate& x, const Coordinate& y, const Coordinate& z,
            const Coordinate& total_site, const RealD a, const Int tag) except +

cdef extern from "qlat/hlbl-contract.h" namespace "qlat":

    cdef cppclass SlTable:
        Int s_limit
        Int l_limit
        std_vector[ComplexD] table
        void init() except +
        void init(const Int s, const Int l) except +
        void init(const Coordinate& total_site) except +
    cdef cppclass CurrentMoments[T]:
        void init() except +
        void init(const Int lsize) except +
    #
    void set_m_z_field_tag(SelectedPoints[RealD]& smf_d, const PointsSelection& psel_d, const Geometry& geo, const Coordinate& xg_x, const Coordinate& xg_y, const RealD a, const Int tag) except +
    #
    void set_local_current_from_props(SelectedPoints[WilsonMatrix]& scf, const SelectedPoints[WilsonMatrix]& sprop1, const SelectedPoints[WilsonMatrix]& sprop2, const PointsSelection& psel_d, const Geometry& geo) except +
    #
    RealD set_psel_d_prob_xy(SelectedPoints[RealD]& psel_d_prob_xy, const PointsSelection& psel, SelectedPoints[RealD]& psel_prob, const PointsSelection& psel_d, const SelectedPoints[RealD]& psel_d_prob, const Long idx_xg_x, const Long idx_xg_y) except +
    #
    void set_current_moments_from_current(CurrentMoments[WilsonMatrix]& cm, const SelectedPoints[WilsonMatrix]& current, const PointsSelection& psel_d, const SelectedPoints[RealD]& psel_d_prob_xy, const Geometry& geo) except +
    #
    void glb_sum_current_moments(CurrentMoments[WilsonMatrix]& cm) except +
    #
    std_vector[std_string] contract_four_pair_labels(const std_vector[std_string]& tags) except +
    #
    std_vector[SlTable] contract_four_pair_no_glb_sum(const ComplexD& coef, const PointsSelection& psel, const PointsSelection& psel_d, const SelectedPoints[RealD]& psel_d_prob_xy, const Geometry& geo, const Long idx_xg_x, const Long idx_xg_y, const SelectedPoints[RealD]& smf_d, const SelectedPoints[WilsonMatrix]& sc_xy, const SelectedPoints[WilsonMatrix]& sc_yx, const CurrentMoments[WilsonMatrix]& cm_xy, const CurrentMoments[WilsonMatrix]& cm_yx, const Int inv_type, const std_vector[std_string]& tags, const Long r_sq_limit, const RealD muon_mass, const RealD z_v) except +
    #
    std_vector[std_string] contract_two_plus_two_pair_labels() except +
    #
    std_vector[SlTable] contract_two_plus_two_pair_no_glb_sum(Long& n_points_selected, Long& n_points_computed, const ComplexD& coef, const Geometry& geo, const PointsSelection& psel, const SelectedPoints[RealD]& psel_prob, const PointsSelection& psel_lps, const SelectedPoints[RealD]& psel_lps_prob, const Long idx_xg_x, const SelectedPoints[ComplexD]& lps_hvp_x, const SelectedPoints[ComplexD]& edl_list_c, const Long r_sq_limit, const RealD muon_mass, const RealD z_v) except +
    #

cdef extern from "qlat/qed.h" namespace "qlat":

    cdef cppclass SpinProp(Field[SpinMatrix]):
        pass
    #
    void free_invert(
        SpinProp& sp_sol, const SpinProp& sp_src,
        const RealD mass, const RealD m5,
        const CoordinateD& momtwist) except +
    #
    void invert_qed(
        SpinProp& sp_sol, const SpinProp& sp_src, const Field[ComplexD]& gf1,
        const RealD mass, const RealD m5, const Int ls,
        const vector[ComplexD]& t_wick_phase_factor_vec,
        const bool is_dagger,
        const RealD stop_rsd, const Long max_num_iter) except +
    #
    Long invert_dwf_qed(
        Field[ComplexD]& f_out4d, const Field[ComplexD]& f_in4d,
        const Field[ComplexD]& gf1, const RealD mass,
        const RealD m5, const Int ls,
        const vector[ComplexD]& t_wick_phase_factor_vec,
        const bool is_dagger,
        const RealD stop_rsd, const Long max_num_iter) except +
    #
    Long cg_with_m_dwf_qed(
        Field[ComplexD]& f_out5d, const Field[ComplexD]& f_in5d,
        const Field[ComplexD]& gf1, const RealD mass,
        const RealD m5, const Int ls,
        const vector[ComplexD]& t_wick_phase_factor_vec,
        const bool is_dagger,
        const RealD stop_rsd, const Long max_num_iter) except +
    #
    void multiply_m_dwf_qed(
        Field[ComplexD]& f_out, const Field[ComplexD]& f_in,
        const Field[ComplexD]& gf1,
        const RealD mass, const RealD m5, const Int ls,
        const vector[ComplexD]& t_wick_phase_factor_vec,
        const bool is_dagger) except +




r"""

# Flow and lattice scale determination\n
Generalize standard Wilson flow to spatial flow.\n
Also, we can define additional flow observable ($W_2(t)$).\n
$$
\ba
W_0(t) =& t^2 \la E(t) \ra
\\
W_1(t) =& W(t) = t \frac{d}{dt} (t^2 \la E(t) \ra)
\\
W_2(t) =& -t^3 \frac{d}{dt}\la E(t) \ra
\ea
$$\n
https://arxiv.org/abs/1006.4518 for $W_0(t)$.
https://arxiv.org/abs/1203.4469 for $W_1(t)$.
"""

### --------------------------------------

import numpy as np

class q:

    from qlat_utils import (
            timer,
            displayln_info,
            TimerFork,
            timer_display,
            set_verbose_level,
            cache_call,
            get_fname,
            does_file_exist_qar_sync_node,
            json_results_append,
            save_pickle_obj,
            RngState,
            is_test,
            get_data_sig_arr,
            Coordinate,
            )

    from qlat.c import (
            FieldRealD,
            GaugeField,
            Geometry,
            gf_plaq_flow_force,
            gf_plaq_field,
            gf_energy_density_dir_field,
            gf_evolve,
            )

### --------------------------------------

is_spatial_default = False
t_dir_default = 3
integrator_type_default = "runge-kutta"
flow_time_default = 0.0

@q.cache_call(maxsize=8)
@q.timer
def get_plaq_factor_for_gf_scale_flow(total_site, is_spatial, t_dir):
    """
    return plaq_factor
    `plaq_factor` can be used in `gf_plaq_flow_force`.
    ```
    plaq_factor[:].shape == (local_volume, 6,)
    ```
    The 6 elements of plaq_factor represent plaq in directions:
    0:xy, 1:xz, 2:xt, 3:yz, 4:yt, 5:zt
    """
    assert isinstance(total_site, tuple)
    geo = q.Geometry(q.Coordinate(total_site))
    plaq_factor = q.FieldRealD(geo, 6)
    if is_spatial:
        plaq_factor.set_zero()
        pfv = plaq_factor[:]
        if t_dir == 3:
            # t dir
            pfv[:, 0] = 1
            pfv[:, 1] = 1
            pfv[:, 3] = 1
        elif t_dir == 0:
            # z dir
            pfv[:, 3] = 1
            pfv[:, 4] = 1
            pfv[:, 5] = 1
        elif t_dir == 1:
            # y dir
            pfv[:, 1] = 1
            pfv[:, 2] = 1
            pfv[:, 5] = 1
        elif t_dir == 2:
            # z dir
            pfv[:, 0] = 1
            pfv[:, 2] = 1
            pfv[:, 4] = 1
        else:
            assert False
    else:
        plaq_factor.set_unit()
    return plaq_factor

@q.timer
def gf_flow_scale(gf, step_size, *, is_spatial=None, t_dir=None, integrator_type=None):
    """
    Modify `gf` in place.
    Default Wilson flow with Runge-Kutta integrator.
    ```
    is_spatial in [ False, True, ]
    t_dir in [ 0, 1, 2, 3, ]
    integrator_type in [ "euler", "runge-kutta", ]
    ```
    Runge-Kutta scheme: http://arxiv.org/abs/1006.4518v3
    """
    fname = q.get_fname()
    if is_spatial is None:
        is_spatial = is_spatial_default
    if t_dir is None:
        t_dir = t_dir_default
    if integrator_type is None:
        integrator_type = integrator_type_default
    geo = gf.geo
    plaq_factor = get_plaq_factor_for_gf_scale_flow(geo.total_site.to_tuple(), is_spatial, t_dir)
    if integrator_type == "euler":
        gm_force = q.gf_plaq_flow_force(gf, plaq_factor)
        q.gf_evolve(gf, gm_force, step_size)
    elif integrator_type == "runge-kutta":
        z = q.gf_plaq_flow_force(gf, plaq_factor)
        z *= 1.0 / 4.0
        q.gf_evolve(gf, z, step_size)
        z *= 17.0 / 9.0
        zp = q.gf_plaq_flow_force(gf, plaq_factor)
        zp *= 8.0 / 9.0
        zp -= z
        q.gf_evolve(gf, zp, step_size)
        z = q.gf_plaq_flow_force(gf, plaq_factor)
        z *= 3.0 / 4.0
        z -= zp
        q.gf_evolve(gf, z, step_size)
    else:
        raise Exception(f"{fname}: integrator_type={integrator_type}")

@q.timer
def gf_plaq_tslice(gf, *, t_dir=None):
    """
    return plaq_arr
    ```
    t_size == geo.total_site[t_dir]
    plaq_arr.shape == (t_size, 6,)
    ```
    """
    if t_dir is None:
        t_dir = t_dir_default
    geo = gf.geo
    total_volume = geo.total_volume
    t_size = geo.total_site[t_dir]
    f_plaq = q.gf_plaq_field(gf)
    plaq_arr = f_plaq.glb_sum_tslice(t_dir=t_dir)[:] / (total_volume / t_size)
    plaq_arr = np.array(plaq_arr, dtype=np.float64)
    return plaq_arr

@q.timer
def gf_energy_density_dir_tslice(gf, *, t_dir=None):
    r"""
    return energy_density_dir_arr
    ```
    t_size == geo.total_site[t_dir]
    energy_density_dir_arr.shape == (t_size, 6,)
    ```
    Similar to `gf_plaq_tslice`
    `gf_energy_density_dir_tslice(gf)[:]` \approx `6 * (1 - gf_plaq_tslice(gf)[:])`
    """
    if t_dir is None:
        t_dir = t_dir_default
    geo = gf.geo
    total_volume = geo.total_volume
    t_size = geo.total_site[t_dir]
    f_edd = q.gf_energy_density_dir_field(gf)
    energy_density_dir_arr = f_edd.glb_sum_tslice(t_dir=t_dir)[:] / (total_volume / t_size)
    energy_density_dir_arr = np.array(energy_density_dir_arr, dtype=np.float64)
    return energy_density_dir_arr

@q.timer(is_timer_fork=True)
def gf_flow_record(
        gf, step_size, num_step,
        *,
        flow_time=None,
        is_spatial=None,
        t_dir=None,
        integrator_type=None,
        ):
    """
    return obj_record 
    obj_record = dict(
        info_list=[
            dict(flow_time, plaq_tslice, energy_density_dir_tslice,),
            ...
        ],
        params=dict(step_size, num_step, flow_time, is_spatial, t_dir, integrator_type,),
    )
    Modify `gf` in place.
    Default Wilson flow with Runge-Kutta integrator.
    ```
    is_spatial in [ False, True, ]
    t_dir in [ 0, 1, 2, 3, ]
    integrator_type in [ "euler", "runge-kutta", ]
    ```
    """
    if flow_time is None:
        flow_time = flow_time_default
    params = dict(
            step_size=step_size,
            num_step=num_step,
            flow_time=flow_time,
            is_spatial=is_spatial,
            t_dir=t_dir,
            integrator_type=integrator_type,
            )
    info_list = []
    for i in range(num_step):
        q.gf_flow_scale(
                gf, step_size,
                is_spatial=is_spatial, t_dir=t_dir,
                integrator_type=integrator_type,
                )
        gf.unitarize()
        flow_time += step_size
        plaq_tslice = gf_plaq_tslice(gf)
        energy_density_dir_tslice = gf_energy_density_dir_tslice(gf)
        info = dict(
            flow_time=flow_time,
            plaq_tslice=plaq_tslice,
            energy_density_dir_tslice=energy_density_dir_tslice,
        )
        info_list.append(info)
    obj_record = dict(
            info_list=info_list,
            params=params,
            )
    return obj_record

default_run_flow_scale_params = dict(
        step_size=0.05,
        num_step=400,
        is_spatial=False,
        t_dir=3,
        integrator_type="runge-kutta",
        )

@q.timer(is_timer_fork=True)
def run_flow_scale(fn_out, *, get_gf=None, fn_gf=None, params=None):
    fname = q.get_fname()
    if not fn_out.endswith(".pickle"):
        q.displayln_info(-1, f"{fname}: WARNING: '{fn_out}' does not endswith '.pickle'. Skip this file.")
        return
    if q.does_file_exist_qar_sync_node(fn_out):
        q.displayln_info(-1, f"{fname}: WARNING: '{fn_out}' for '{fn_gf}' already exist.")
        return
    if get_gf is None:
        if not q.does_file_exist_qar_sync_node(fn_gf):
            q.displayln_info(-1, f"{fname}: WARNING: '{fn_gf}' does not exist. Skip this file.")
            return
    else:
        assert fn_gf is None
    if params is None:
        params = q.default_run_flow_scale_params.copy()
    else:
        params = q.default_run_flow_scale_params | params
    q.json_results_append(f"{fname}: Start compute flow scale info fn='{fn_out}' for '{fn_gf}'")
    if get_gf is None:
        gf = q.GaugeField()
        gf.load(fn_gf)
    else:
        gf = get_gf().copy()
    step_size = params["step_size"]
    num_step = params["num_step"]
    is_spatial = params["is_spatial"]
    t_dir = params["t_dir"]
    integrator_type = params["integrator_type"]
    obj_record = q.gf_flow_record(
            gf, step_size, num_step,
            is_spatial=is_spatial, t_dir=t_dir, integrator_type=integrator_type)
    q.save_pickle_obj(obj_record, fn_out)
    if q.is_test():
        q.json_results_append(f"{fname}: params={obj_record['params']}")
        q.json_results_append(f"{fname}: flow_time", obj_record['info_list'][-1]['flow_time'])
        rs = q.RngState()
        q.json_results_append(
                f"{fname}: sig(plaq_tslice)",
                q.get_data_sig_arr(obj_record['info_list'][-1]['plaq_tslice'], rs, 3))
        q.json_results_append(
                f"{fname}: sig(energy_density_dir_tslice)",
                q.get_data_sig_arr(obj_record['info_list'][-1]['energy_density_dir_tslice'], rs, 3))
    q.json_results_append(f"{fname}: End compute flow scale info fn='{fn_out}' for '{fn_gf}'")

### --------------------------------------

(
    q.get_plaq_factor_for_gf_scale_flow,
    q.gf_flow_scale,
    q.gf_plaq_tslice,
    q.gf_energy_density_dir_tslice,
    q.gf_flow_record,
    q.run_flow_scale,
    q.default_run_flow_scale_params,
) = (
    get_plaq_factor_for_gf_scale_flow,
    gf_flow_scale,
    gf_plaq_tslice,
    gf_energy_density_dir_tslice,
    gf_flow_record,
    run_flow_scale,
    default_run_flow_scale_params,
)
